#
# Copyright (c) 2022 Arm Limited
# Copyright (c) 2022 Hanno Becker
# Copyright (c) 2023 Amin Abdulrahman, Matthias Kannwischer
# SPDX-License-Identifier: MIT
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Author: Hanno Becker <hannobecker@posteo.de>
#

"""
SLOTHY configuration
"""
import shutil

from copy import deepcopy

from slothy.helper import LockAttributes, NestedPrint


class InvalidConfig(Exception):
    """Exception raised when an invalid SLOTHY configuration is detected"""


class Config(NestedPrint, LockAttributes):
    """Configuration for Slothy.

    This configuration object is used both for one-shot optimizations using
    SlothyBase, as well as stateful multi-pass optimizations using Slothy.
    """

    @property
    def arch(self):
        """The module defining the underlying architecture used by Slothy.

        TODO: Add details on what exactly is assumed about this module.
        """
        return self._arch

    @property
    def target(self):
        """The module defining the target microarchitecture used by Slothy.

        TODO: Add details on what exactly is assumed about this module.
        """
        return self._target

    @property
    def outputs(self):
        """List defining of architectural or symbolic registers that should
        be considered as outputs of the input snippet.
        """
        return self._outputs

    @property
    def reserved_regs(self):
        """Set of architectural registers _not_ available for register renaming.
        May be unset (None) to pick the default reserved registers for the target
        architecture.

        In the lingo of inline assembly, this can be seen as the complement of
        the clobber list.

        .. note::

            Reserved registers are, by default, considered  "locked": They
            will not be _introduced_ during renaming, but existing uses will not
            be touched. If you want to remove existing uses of reserved registers
            through renaming, you should disable `reserved_regs_are_locked`.

        .. warning::

            When this is set, it _overwrites_ the default reserved registers for
            the target architecture. If you still want the default reserved
            registers to remain reserved, you have to explicitly list them!
        """
        if self._reserved_regs is not None:
            return self._reserved_regs
        return self._arch.RegisterType.default_reserved()

    @property
    def reserved_regs_are_locked(self):
        """Indicates whether reserved registers should be locked by default.

        Reserved registers are not introduced during renaming. However, where
        they are already used by the input assembly, their use will not be
        eliminated or altered -- that is, reserved registers are 'locked' by
        default.

        Disable this configuration option to allow (in fact, force) renaming
        of existing uses of reserved registers. This can be useful when trying
        to eliminate uses of particular registers from some piece of assembly.
        """
        return self._reserved_regs_are_locked

    @property
    def selftest(self):
        """
        Indicates whether SLOTHY performs an empirical equivalence-test on the
        optimization results.

        When this is set, and if the target architecture and host platform support it,
        this will run an empirical equivalence checker trying to confirm that the
        input and output of SLOTHY are likely functionally equivalent.

        The primary purpose of this checker is to detect issue that would presently
        be overlooked by the selfcheck:

        * The selfcheck is currently blind to address offset fixup. If something goes
          wrong, the input and output will not be functionally equivalent, but we would
          only notice once we actually compile and run the code. The selftest will
          likely catch issues.

        * When using software pipelining, the selfcheck reduces to a straightline check
          for a bounded unrolling of the loop. An unbounded selfcheck is currently not
          implemented. With the selftest, you still need to fix a loop bound, but at
          least you can equivalence-check the loop-form (including the compare+branch
          instructions at the loop boundary) rather than the unrolled code.

        .. important::

            To run this, you need `llvm-nm`, `llvm-readobj`, `llvm-mc`
            in your PATH. Those are part of a standard LLVM setup.

        .. note::

            This is so far implemented as a repeated randomized test -- nothing clever.
        """
        selftest_enabled = self._selftest
        if None in [
            shutil.which(tool) for tool in ["llvm-mc", "llvm-nm", "llvm-readobj"]
        ]:
            self.logger.warning(
                (
                    "LLVM (llvm-mc, llvm-nm, llvm-readobj) not found, "
                    "disabling selftest. Consider installing LLVM."
                )
            )
            selftest_enabled = False

        return selftest_enabled

    @property
    def selftest_iterations(self):
        """If selftest is set, indicates the number of random selftest to conduct"""
        return self._selftest_iterations

    @property
    def selftest_address_registers(self):
        """Dictionary of (reg, sz) items indicating which registers are assumed to be
        pointers to memory, and if so, of what size.
        """
        return self._selftest_address_registers

    @property
    def selftest_default_memory_size(self):
        """Default buffer size to use for registers which are automatically inferred to be
        used as pointers and for which no memory size has been configured via
        `address_registers`.
        """
        return self._selftest_default_memory_size

    @property
    def selfcheck(self):
        """Indicates whether SLOTHY performs a self-check on the optimization result.

        The selfcheck confirms that the scheduling permutation found by SLOTHY yields
        an isomorphism between the data flow graphs of the original and optimized code.

        .. warning::

            Do not unset this option unless you know what you are doing.
            It is vital in catching bugs in the model generation early.

        .. warning::

            The selfcheck is not a formal verification of SLOTHY's output!
            There are at least two classes of bugs uncaught by the selfcheck:

            * User configuration issues: The selfcheck validates SLOTHY's optimization
              in the context of the provided configuration. Validation of the
              configuration is the user's responsibility. Two common pitfalls include
              missing reserved registers (allowing SLOTHY to clobber more registers than
              intended), or missing output registers (allowing SLOTHY to overwrite an
              output register in subsequent instructions).

              This is the most common source of issues for code passing the selfcheck
              but remaining functionally incorrect.

            * Bugs in address offset fixup: SLOTHY's modelling of post-load/store address
              increments is deliberately inaccurate to allow for reordering of such
              instructions leveraging commutativity relations such as

            .. code-block:: asm

                LDR X,[A],#imm;  STR Y,[A]    ===     STR Y,[A, #imm];  LDR X,[A],#imm


            .. hint::

                See also section "Address offset rewrites" in the SLOTHY paper

            Bugs in SLOTHY's address fixup logic would not be caught by the selfcheck.
            If your code doesn't work and you are sure to have configured SLOTHY
            correctly, you may therefore want to double-check that address offsets have
            been adjusted correctly by SLOTHY.
        """
        return self._selfcheck

    @property
    def selfcheck_failure_logfile(self):
        """The filename for the log of a failing selfcheck.

        This is printed in the terminal as well, but difficult to analyze for its
        sheer size.
        """
        return self._selfcheck_failure_logfile

    @property
    def unsafe_address_offset_fixup(self):
        """Whether address offset fixup is enabled

        Address offset fixup is a feature which leverages commutativity relations
        such as

        .. code-block:: asm

            ldr X, [A], #immA;
            str Y, [A, #immB]
            ==
            str Y, [A, #(immB+immA)]
            ldr X, [A], #immA

        to achieve greater instruction scheduling flexibility in SLOTHY.

        .. important::

            When you enable this feature, you MUST ensure that registers which are
            used for addresses are not used in any other instruction than load and
            stores. OTHERWISE, THE USE OF THIS FEATURE IS UNSOUND (you may see ldr/
            str instructions with increment reordered with instructions depending
            on the address register).

        By default, this is enabled for backwards compatibility.

        .. note::

            The user-imposed safety constraint is not a necessity -- in principle,
            SLOTHY could detect when it is safe to reorder ldr/str instructions with
            increment.
            It just hasn't been implemented yet.
        """
        return self._unsafe_address_offset_fixup

    @property
    def allow_useless_instructions(self):
        """Indicates whether SLOTHY should abort upon encountering unused instructions.

        SLOTHY requires explicit knowledge of the intended output registers of its
        input assembly. If this option is set, and an instruction is encountered which
        writes to a register which (a) is not an output register, (b) is not used by
        any later instruction, then SLOTHY will flag this instruction and abort.

        The reason for this behaviour is that such unused instructions are usually
        a sign of a buggy configuration, which would likely lead to intended output
        registers being clobbered by later instructions.

        .. warning::

            Don't disable this option unless you know what you are doing!
            Disabling this option makes it much easier to overlook configuration
            issues in SLOTHY and can lead to hard-to-debug optimization failures.
        """
        return self._allow_useless_instructions

    @property
    def variable_size(self):
        """Model number of stalls as a parameter in the constraint model.

        If this is set, one-shot SLOTHY optimization will make the number of stalls
        flexible in the model and, by default, task the underlying constraint solver
        to minimize it.

        If this is not set, one-shot SLOTHY optimizations will search for solutions
        with a fixed number of stalls, and an external binary search be used to
        find the minimum number of stalls.

        For small-to-medium sizes assembly input, this option should be set, and will
        lead to faster optimization. For large assembly input, the user should experiment
        and consider unsetting it to reduce model complexity.
        """
        return self._variable_size

    @property
    def keep_tags(self):
        """Indicates whether tags in the input source should be kept or removed.

        Tags include pre/core/post or ordering annotations that usually become meaningless
        post-optimization. However, for preprocessing runs that do not reorder code, it
        makes sense to keep them.
        """
        return self._keep_tags

    @property
    def inherit_macro_comments(self):
        """Indicates whether comments at macro invocations should be inherited to
        instructions in the macro body.
        """
        return self._inherit_macro_comments

    @property
    def ignore_tags(self):
        """Indicates whether tags in the input source should be ignored."""
        return self._ignore_tags

    @property
    def register_aliases(self):
        """Dictionary mapping symbolic register names to architectural register names.
        When using Slothy, this can be indirectly populated by placing `.req` expressions
        in the input assembly. When using SlothyBase directly, this needs to be filled
        in by hand.

        This is always joined with a list of default aliases (such as lr mapping to r14)
        specified in the target architecture.
        """
        return {**self._register_aliases, **self._arch.RegisterType.default_aliases()}

    def add_aliases(self, new_aliases):
        """Add further register aliases to the configuration"""
        self._register_aliases = {**self._register_aliases, **new_aliases}

    @property
    def rename_inputs(self):
        """A dictionary mapping input register names (symbolic or architectural) to their
        renaming configuration.

        There are three supported renaming configurations per input: "static", "any", or
        a fixed architectural register. The configuration "any" means that the input may
        be freely renamed, and that the renaming is chosen at model solving time. This is
        the most flexible, but also the most demanding option. The configuration "static"
        means that the renaming is chosen at model construction time, as follows: If the
        input does already have an architectural name, it will keep it. If, otherwise, it
        is a symbolic input, it will be statically assigned an architectural name at model
        construction time. Finally, if the input is explicitly assigned an architectural
        register name, this name will be enforced. Note that this even applies to inputs
        which already have an architectural name -- that is, you can use this option to
        change the architectural allocation of inputs.

        The special keys "symbolic", "arch" apply to all symbolic and architectural
        inputs, respectively. The key "other" applies to all inputs for which no other
        key matches.

        The default value is { "symbolic": "any", "arch" : "static" } -- that is,
        architectural inputs are not renamed, while symbolic inputs are dynamically
        renamed.

        Examples:

        * Generally, unless you are prepared to modify surrounding code, you should
          have "arch" : "static", which will not rename inputs which already have
          architectural register names.
        * Config.rename_inputs = { "other" : "any" }
          This would rename _all_ inputs, regardless of whether they're symbolic or not.
          Thus, you'd likely need to modify surrounding code.
        * Config.rename_inputs = { "in" : "r0", "arch" : "static", "symbolic" : "any" }
          This would rename the symbolic input GPR 'in' to 'r0', keep all other inputs
          which already have an architectural name, while dynamically assigning suitable
          registers for symbolic inputs.

        In case of a successful optimization, the assignment of input registers to
        architectural registers is given by the dictionary Result.input_renaming.
        """
        self._check_rename_config(self._rename_inputs)
        return self._rename_inputs

    @property
    def rename_outputs(self):
        """A dictionary mapping output register names (symbolic or architectural) to their
        renaming configuration.

        Analogous to Config.rename_inputs.

        The default value is { "symbolic": "any", "arch" : "static" } -- that is,
        architectural outputs are not renamed, while symbolic outputs are dynamically
        renamed.

        In case of a successful optimization, the assignment of input registers to
        architectural registers is given by the dictionary Result.input_renaming.
        """
        self._check_rename_config(self._rename_outputs)
        return self._rename_outputs

    @property
    def inputs_are_outputs(self):
        """If set, any input in the assembly to be optimized (that is, every register
        that is used as an input before it has been written to) is treated as an output.
        _Moreover_, such simultaneous input-outputs are forced to reside in the same
        architectural register at the beginning and end of the snippet.

        This should usually be set when optimizing loops.
        """
        return self._inputs_are_outputs

    @property
    def locked_registers(self):
        """List of architectural registers that should not be renamed when they are
        used as output registers. Reserved registers are treated as locked if
        the option `reserved_regs_are_locked` is set.
        """
        if self.reserved_regs_are_locked:
            return set(self.reserved_regs).union(self._locked_registers)
        else:
            return set(self._locked_registers)

    @property
    def sw_pipelining(self):
        """Subconfiguration for software pipelining. Enabled/Disabled
        via the sub-field sw_pipelining.enabled. See Config.SoftwarePipelining
        for more information.
        """
        return self._sw_pipelining

    @property
    def constraints(self):
        """Subconfiguration for constraints to be considered by SLOTHY,
        e.g. whether latencies or functional units are modelled.
        See Config.Constraints for more information.
        """
        return self._constraints

    @property
    def hints(self):
        """Subconfiguration for hints to be considered by SLOTHY.
        See Config.Hints for more information.
        """
        return self._hints

    @property
    def max_solutions(self):
        """The maximum number of solution found by the underlying constraint
        solver before it stops the search.
        """
        return self._max_solutions

    @property
    def with_preprocessor(self):
        """Indicates whether the C preprocessor is run prior to optimization."""
        return self._with_preprocessor

    @property
    def with_llvm_mca(self):
        """Indicates whether LLVM MCA should be run prior and after optimization
        to obtain approximate performance data based on LLVM's scheduling models.

        If this is set, Config.compiler_binary need to be set, and llcm-mca in
        your PATH.
        """
        return self._with_llvm_mca_before and self._with_llvm_mca_after

    @property
    def llvm_mca_full(self):
        """Indicates whether all available statistics from LLVM MCA should be printed."""
        return self._llvm_mca_full

    @property
    def llvm_mca_issue_width_overwrite(self):
        """Overwrite LLVM MCA's in-built issue width with the one SLOTHY uses"""
        return self._llvm_mca_issue_width_overwrite

    @property
    def with_llvm_mca_before(self):
        """Indicates whether LLVM MCA should be run prior to optimization
        to obtain approximate performance data based on LLVM's scheduling models.

        If this is set, Config.compiler_binary need to be set, and llcm-mca in
        your PATH.
        """
        return self._with_llvm_mca_before

    @property
    def with_llvm_mca_after(self):
        """Indicates whether LLVM MCA should be run after optimization
        to obtain approximate performance data based on LLVM's scheduling models.

        If this is set, Config.compiler_binary need to be set, and llcm-mca in
        your PATH.
        """
        return self._with_llvm_mca_after

    @property
    def compiler_binary(self):
        """The compiler binary to be used.

        This is only relevant if `with_preprocessor` or `with_llvm_mca_before`
        or `with_llvm_mca_after` are set."""
        return self._compiler_binary

    @property
    def compiler_include_paths(self):
        """Include path to add to compiler invocations

        This is only relevant if `with_preprocessor` or `with_llvm_mca_before`
        or `with_llvm_mca_after` are set."""
        return self._compiler_include_paths

    @property
    def timeout(self):
        """The timeout in seconds after which each invocation of the underlying
        constraint solver stops its search. A positive integer."""
        return self._timeout

    @property
    def retry_timeout(self):
        """The timeout in seconds after which the underlying constraint solver stops
        its search, in case of secondary optimization passes for other objectives than
        performance optimization (e.g., minimization of iteration overlapping)."""
        return self._retry_timeout

    @property
    def do_address_fixup(self):
        """Indicates whether post-optimization address fixup should be conducted.

        SLOTHY's modelling of post-load/store address increments is deliberately
        inaccurate to allow for reordering of such instructions leveraging commutativity
        relations such as:

        ```
        LDR X,[A],#imm;  STR Y,[A]    ===     STR Y,[A, #imm];  LDR X,[A],#imm
        ```

        When such reordering happens, a "post-optimization address fixup" of immediate
        load/store offsets is necessary. See also section "Address offset rewrites" in
        the SLOTHY paper.

        Disabling this option will skip post-optimization address fixup and put the
        burden of post-optimization address fixup on the user.
        Disabling this option does NOT tighten the constraint model to forbid reorderings
        such as the above.

        WARNING: Don't disable this option unless you know what you are doing!
            Disabling this will likely lead to optimized code that is functionally
            incorrect and needing manual address offset fixup!
        """
        return self._do_address_fixup

    @property
    def ignore_objective(self):
        """Indicates whether the secondary objective (such as minimization of iteration
        overlapping) should be ignored."""
        return self._ignore_objective

    @property
    def objective_precision(self):
        """The proximity to the estimated optimum solution at which the solver will
        stop its search.

        For example, a value of 0.05 means that the solver will stop when the current
        solution is within 5% of the current estimate for the optimal solution."""
        return self._objective_precision

    @property
    def objective_lower_bound(self):
        """A lower bound for the objective at which to stop the search."""
        return self._objective_lower_bound

    @property
    def has_objective(self):
        """Indicates whether a different objective than minimization of stalls
        has been registered."""
        objectives = sum(
            [
                self.sw_pipelining.enabled
                and self.sw_pipelining.minimize_overlapping is True,
                self.constraints.maximize_register_lifetimes is True,
                self.constraints.minimize_spills is True,
                self.constraints.move_stalls_to_top is True,
                self.constraints.move_stalls_to_bottom is True,
                self.constraints.minimize_register_usage is not None,
                self.constraints.minimize_use_of_extra_registers is not None,
                self.target.has_min_max_objective(self),
            ]
        )
        if objectives > 1:
            raise InvalidConfig("Can only pick one optimization objective")

        return objectives == 1

    @property
    def absorb_spills(self):
        return self._absorb_spills

    @property
    def split_heuristic(self):
        """Trade-off between runtime and optimality: Split each code block
        to be optimized into a fixed number of subchunks and optimize them
        one by one, rather than attempting a single large optimization.

        If enabled, the numeric option split_heuristic_factor determines the
        number of factors to split each block of code into.
        """
        return self._split_heuristic

    @property
    def split_heuristic_factor(self):
        """If split_heuristic is enabled, the number of factors to split each
        code block into prior to passing it to the core of Slothy.

        The value of this option is irrelevant if split_heuristic is False.
        """
        if not self.split_heuristic:
            raise InvalidConfig(
                "Did you forget to set config.split_heuristic=True? "
                "Shouldn't read config.split_heuristic_factor otherwise."
            )
        return self._split_heuristic_factor

    @property
    def split_heuristic_abort_cycle_at_high(self):
        """During the split heuristic, a threshold for the number of stalls in the current
        optimization window above which the current pass of the split heuristic should
        stop.
        """
        if not self.split_heuristic:
            raise InvalidConfig(
                "Did you forget to set config.split_heuristic=True? "
                "Shouldn't read config.split_heuristic_abort_cycle_at otherwise."
            )
        return self._split_heuristic_abort_cycle_at_high

    @property
    def split_heuristic_abort_cycle_at_low(self):
        """During the split heuristic, a threshold for the number of stalls in the current
        optimization window below which the current pass of the split heuristic should
        stop.
        """
        if not self.split_heuristic:
            raise InvalidConfig(
                "Did you forget to set config.split_heuristic=True? "
                "Shouldn't read config.split_heuristic_abort_cycle_at otherwise."
            )
        return self._split_heuristic_abort_cycle_at_low

    @property
    def split_heuristic_stepsize(self):
        """If split heuristic is used, the increment for the sliding window. By default,
        this is twice the split factor. For example, a split factor of 5 means that the
        window size is 0.2 of the overall code size, and the default step size of 0.1
        means that the sliding windows will be [0,0.2], [0.1,0.3], ..."""
        if not self.split_heuristic:
            raise InvalidConfig(
                "Did you forget to set config.split_heuristic=True? "
                "Shouldn't read config.split_heuristic_stepsize otherwise."
            )
        return self._split_heuristic_stepsize

    @property
    def split_heuristic_optimize_seam(self):
        """If the split heuristic is used, the number of instructions above and beyond
        the current sliding window that should be fixed but taken into account during
        optimization."""
        if not self.split_heuristic:
            raise InvalidConfig(
                "Did you forget to set config.split_heuristic=True? "
                "Shouldn't read config.split_heuristic_optimize_seam otherwise."
            )
        return self._split_heuristic_optimize_seam

    @property
    def split_heuristic_chunks(self):
        """If split heuristic is used, explicitly lists the optimization windows to be
        used. If unset, a sliding or adaptive optimization window will be used."""
        if not self.split_heuristic:
            raise InvalidConfig(
                "Did you forget to set config.split_heuristic=True? "
                "Shouldn't read config.split_heuristic_chunks otherwise."
            )
        return self._split_heuristic_chunks

    @property
    def split_heuristic_bottom_to_top(self):
        """If the split heuristic is used, move the sliding window from bottom to top
        rather than from top to bottom."""
        if not self.split_heuristic:
            raise InvalidConfig(
                "Did you forget to set config.split_heuristic=True? "
                "Shouldn't read config.split_heuristic_bottom_to_top otherwise."
            )
        return self._split_heuristic_bottom_to_top

    @property
    def split_heuristic_region(self):
        """Restrict the split heuristic to a sub-region of the code.

        For example, if this is set to [0.25,0.75], only the middle half of the input will
        be optimized through the split heuristic.

        This option can be combined with other options such as the split factor.
        For example, if the split region is set fo [0.25, 0.75] and the split factor is
        5, then optimization windows of size .1 will be considered within [0.25, 0.75].

        Note that even if this option is used, the specification of inputs and outputs is
        still with respect to the entire code; SLOTHY will automatically derive the
        outputs of the subregion configured here."""
        if not self.split_heuristic:
            raise InvalidConfig(
                "Did you forget to set config.split_heuristic=True? "
                "Shouldn't read config.split_heuristic_region otherwise."
            )
        return self._split_heuristic_region

    @property
    def split_heuristic_preprocess_naive_interleaving(self):
        """Prior to applying the split heuristic, interleave instructions according
        to lowest depth, without applying register renaming.

        This can be useful if the code to be optimized is comprised of independent
        computations operating on different architectural state (e.g. scalar vs. SIMD);
        in this case, the naive preprocessing will 'zip' the different computations prior
        to applying the core optimization."""
        if not self.split_heuristic:
            raise InvalidConfig(
                "Did you forget to set config.split_heuristic=True? "
                "Shouldn't read config.split_heuristic_preprocess_naive_interleaving "
                "otherwise."
            )
        return self._split_heuristic_preprocess_naive_interleaving

    @property
    def split_heuristic_preprocess_naive_interleaving_by_latency(self):
        """If split heuristic with naive preprocessing is used, this option causes
        the naive interleaving to be by latency-depth rather than latency."""
        if not self.split_heuristic:
            raise InvalidConfig(
                "Did you forget to set config.split_heuristic=True? Shouldn't"
                "read config.split_heuristic_preprocess_naive_interleaving_by_latency "
                "otherwise."
            )
        return self._split_heuristic_preprocess_naive_interleaving_by_latency

    @property
    def split_heuristic_estimate_performance(self):
        """After applying the split heuristic, run SLOTHY again on the entire code to
        estimate the performance and display un-used issue slots in the output."""
        if not self.split_heuristic:
            raise InvalidConfig(
                "Did you forget to set config.split_heuristic=True? Shouldn't"
                "read config.split_heuristic_estimate_performance otherwise."
            )
        return self._split_heuristic_estimate_performance

    @property
    def split_heuristic_repeat(self):
        """If split_heuristic is enabled, the number of times the splitting heuristic
        should be repeated.

        Note: This is an experimental option the practical value of which has not yet
        been thoroughly studied. Try if you like, but beware the bump in runtime for
        the optimization.

        The value of this option is irrelevant if split_heuristic is False.
        """
        if not self.split_heuristic:
            raise InvalidConfig(
                "Did you forget to set config.split_heuristic=True? "
                "Shouldn't read config.split_heuristic_repeat otherwise."
            )
        return self._split_heuristic_repeat

    @property
    def split_heuristic_preprocess_naive_interleaving_strategy(self):
        """Strategy for naive interleaving preprocessing step

        Supported values are:

        * "depth": Always pick the instruction with the lower possible
          depth in the DFG first.
        * "alternate": Try to evenly alternate between instructions tagged with
          "interleaving_class=0/1".

        """
        return self._split_heuristic_preprocess_naive_interleaving_strategy

    def copy(self):
        """Make a deep copy of the configuration"""
        # Temporarily unset references to Arch and Target for deepcopy
        arch, target = self.arch, self.target
        self.arch = self.target = None
        res = deepcopy(self)
        res.arch, res.target = arch, target
        self.arch, self.target = arch, target
        return res

    class SoftwarePipelining(NestedPrint, LockAttributes):
        """Subconfiguration for software pipelining"""

        @property
        def enabled(self):
            """Determines whether software pipelining should be enabled."""
            return self._enabled

        @property
        def unroll(self):
            """The number of times the loop body should be unrolled."""
            return self._unroll

        @property
        def pre_before_post(self):
            """If both early and late instructions are allowed, force late instructions
            of iteration N to come _before_ early instructions of iteration N+2."""
            return self._pre_before_post

        @property
        def allow_pre(self):
            """Allow 'early' instructions, that is, instructions that are pulled forward
            from iteration N+1 to iteration N. A typical example would be an early load.
            """
            return self._allow_pre

        @property
        def allow_post(self):
            """Allow 'late' instructions, that is, instructions that are deferred from
            iteration N to iteration N+1. A typical example would be a late store."""
            return self._allow_post

        @property
        def unknown_iteration_count(self):
            """Determines whether the number of iterations is statically known and larger
            than the number of exceptional iterations hoisted out by SLOTHY (at most 2).

            Set this to `True` if the loop can have any number of iterations."""
            return self._unknown_iteration_count

        @property
        def minimize_overlapping(self):
            """Set the objective to minimize the amount of iteration overlapping"""
            return self._minimize_overlapping

        @property
        def optimize_preamble(self):
            """Perform a separate optimization pass for the loop preamble."""
            return self._optimize_preamble

        @property
        def optimize_postamble(self):
            """Perform a separate optimization pass for the loop postamble."""
            return self._optimize_postamble

        @property
        def max_overlapping(self):
            """The maximum number of early or late instructions.
            `None` means that any number of early/late instructions is allowed."""
            return self._max_overlapping

        @property
        def min_overlapping(self):
            """The minimum number of early or late instructions.
            `None` means that any number of early/late instructions is allowed."""
            return self._min_overlapping

        @property
        def halving_heuristic(self):
            """Performance improvement heuristic: Rather than running a
            general software pipelining optimization, proceed in two steps:
            First, optimize loop body _without_ software pipelining. Then,
            split it as [A;B] and optimize [B;A]. The final result is then
            `A; optimized([B;A]); B`, with `A` being the preamble, `B` the
            postamble, and `optimized([B;A])` the loop kernel."""
            return self._halving_heuristic

        @property
        def halving_heuristic_periodic(self):
            """Variant of the halving heuristic: Consider loop boundary when
            optimizing [B;A] in the second step of the halving heuristic.
            This is computationally more expensive but avoids bottlenecks
            at the loop boundary that could otherwise ensue.

            This is only meaningful is the halving heuristic is enabled."""
            return self._halving_heuristic_periodic

        @property
        def halving_heuristic_split_only(self):
            """Cut-down version of halving-heuristic which only splits the loop
            `[A;B]` into `A; [B;A]; B` but does not perform optimizations."""
            return self._halving_heuristic_split_only

        @property
        def max_pre(self):
            """The maximum relative position (between 0 and 1) of an instruction
            that should be considered as a potential early instruction.
            For example, a value of 0.5 means that only instruction in the
            first half of the original loop body are considered as potential
            early instructions."""
            return self._max_pre

        def __init__(self):
            super().__init__()

            self.enabled = False
            self.unroll = 1
            self.pre_before_post = False
            self.allow_pre = True
            self.allow_post = False
            self.unknown_iteration_count = False
            self.minimize_overlapping = True
            self.optimize_preamble = True
            self.optimize_postamble = True
            self.max_overlapping = None
            self.min_overlapping = None
            self.halving_heuristic = False
            self.halving_heuristic_periodic = False
            self.halving_heuristic_split_only = False
            self.max_pre = 1.0

            self.lock()

        @enabled.setter
        def enabled(self, val):
            self._enabled = val

        @unroll.setter
        def unroll(self, val):
            self._unroll = val

        @pre_before_post.setter
        def pre_before_post(self, val):
            self._pre_before_post = val

        @allow_pre.setter
        def allow_pre(self, val):
            self._allow_pre = val

        @allow_post.setter
        def allow_post(self, val):
            self._allow_post = val

        @unknown_iteration_count.setter
        def unknown_iteration_count(self, val):
            self._unknown_iteration_count = val

        @minimize_overlapping.setter
        def minimize_overlapping(self, val):
            self._minimize_overlapping = val

        @optimize_preamble.setter
        def optimize_preamble(self, val):
            self._optimize_preamble = val

        @optimize_postamble.setter
        def optimize_postamble(self, val):
            self._optimize_postamble = val

        @max_overlapping.setter
        def max_overlapping(self, val):
            self._max_overlapping = val

        @min_overlapping.setter
        def min_overlapping(self, val):
            self._min_overlapping = val

        @halving_heuristic.setter
        def halving_heuristic(self, val):
            self._halving_heuristic = val

        @halving_heuristic_periodic.setter
        def halving_heuristic_periodic(self, val):
            self._halving_heuristic_periodic = val

        @halving_heuristic_split_only.setter
        def halving_heuristic_split_only(self, val):
            self._halving_heuristic_split_only = val

        @max_pre.setter
        def max_pre(self, val):
            self._max_pre = val

    class Constraints(NestedPrint, LockAttributes):
        """Subconfiguration for performance constraints"""

        @property
        def stalls_allowed(self):
            """The number of stalls allowed. Internally, this is the number of NOP
            instructions that SLOTHY introduces before attempting to find a stall-free
            version of the code (or, more precisely: a version matching all constraints,
            which may be weaker than stall-free).

            This is only meaningful for direct invocations to SlothyBase. You should not
            set this field when interfacing with Slothy."""
            if self.functional_only:
                return 0
            return self._stalls_allowed

        @property
        def stalls_maximum_attempt(self):
            """The maximum number of stalls to attempt before aborting the optimization
            and reporting it as infeasible.

            Note that since SLOTHY does not (yet?) introduce stack spills, a symbolic
            assembly snippet may be impossible to even concretize with architectural
            register names, regardless of the number of stalls one allows."""
            if self.functional_only:
                return 0
            return self._stalls_maximum_attempt

        @property
        def stalls_minimum_attempt(self):
            """The minimum number of stalls to attempt.

            This may be useful if it's known for external reasons that searching for
            optimiztions with less stalls is infeasible."""
            if self.functional_only:
                return 0
            return self._stalls_minimum_attempt

        @property
        def stalls_first_attempt(self):
            """The first number of stalls to attempt.

            This may be useful if it's known for external reasons that searching for
            optimization with less stalls is infeasible."""
            if self.functional_only:
                return 0
            return self._stalls_first_attempt

        @property
        def stalls_precision(self):
            """The precision of the binary search for the minimum number of stalls

            SLOTHY will stop searching if it can narrow down the minimum number
            of stalls to an interval of the length provided by this variable.
            In particular, a value of 1 means the true minimum if searched for."""
            if self.functional_only:
                return 1
            return self._stalls_precision

        @property
        def stalls_timeout_below_precision(self):
            """If this variable is set to a non-None value, SLOTHY does not abort
            optimization once binary search is operating on an interval smaller than
            the stall precision, but instead sets a different (typically smaller) timeout.
            """
            return self._stalls_timeout_below_precision

        @property
        def model_latencies(self):
            """Determines whether instruction latencies should be modelled.

            When set, SLOTHY will enforce that instructions are placed in accordance
            with the latency of the instructions that they depend on."""
            return self._model_latencies

        @property
        def model_functional_units(self):
            """Determines whether functional units should be modelled.

            When set, SLOTHY will enforce that instructions are placed in accordance
            with the presence and throughput of functional units that they depend on."""
            return self._model_functional_units

        @property
        def functional_only(self):
            """Limit Slothy to register renaming"""
            return (
                self.model_functional_units is False and self.model_latencies is False
            )

        @property
        def allow_reordering(self):
            """Allow Slothy to reorder instructions

            Disabling this may be useful to e.g. reassign register names
            in code that has already been scheduled properly."""
            return self._allow_reordering

        @property
        def allow_renaming(self):
            """Allow Slothy to rename registers

            Disabling this may be useful in conjunction with !allow_reordering
            in order to find the number of model violations in a piece of code."""
            return self._allow_renaming

        @property
        def allow_spills(self):
            """Allow Slothy to introduce stack spills

            When this option is enabled, Slothy will consider the introduction
            of stack spills to reduce register pressure. The naming of stack
            locations can be customized using spill_stack_loc_prefix.

            This option should only be disabled if it is known that the input
            assembly suffers from high register pressure. For example, this can
            be the case for symbolic input assembly."""
            return self._allow_spills

        @property
        def spill_stack_loc_prefix(self):
            """
            Prefix for stack location names in generated assembly when spills
            are enabled.

            When allow_spills is True, this prefix is used to generate unique stack
            location names in the format "{prefix}_{index}" (e.g., "STACK_LOC_0",
            "STACK_LOC_1", etc.). This allows customization of naming conventions
            to match project requirements.

            Default: "STACK_LOC"
            """
            return self._spill_stack_loc_prefix

        @property
        def spill_type(self):
            """The type of spills to generate

            This is usually spilling to the stack, but other options may exist.
            For example, on Armv7-M microcontrollers it can be useful to spill
            from the GPR file to the FPR file.

            It is expected that this option is set as a dictionary, for example,
            with the key determining whether the spills are supposed to be to
            the stack or to the FPR file, and the value defining a starting
            index for the FPRs in the latter case.

            The exact influence of this option is architecture dependent. You
            should consult the `Spill` class in the target architecture model to
            understand the options."""
            if self._spill_type is None:
                return {}
            else:
                return self._spill_type

        @property
        def minimize_spills(self):
            """Minimize number of stack spills

            When this option is enabled, the Slothy will pass minimization of
            stack spills as the optimization objective to the solver.
            """
            return self._minimize_spills

        @property
        def max_displacement(self):
            """The maximum relative displacement of an instruction.

            Examples:

            * If set to 1, instructions can be reordered freely.
            * If set to 0, no reordering will happen.
            * If set to 0.5, an instruction will not move by more than N/2
              places between original and re-scheduled source code.

            This is an experimental feature for the purpose of speeding
            up otherwise intractable optimization tasks.

            .. warning::

                This only takes effect in straightline optimization (no software
                pipelining).
            """
            return self._max_displacement

        def __init__(self):
            super().__init__()

            # TODO: Move those to target specific configuration
            self.st_ld_hazard = True
            self.st_ld_hazard_ignore_scattergather = False
            self.st_ld_hazard_ignore_stack = False
            self.minimize_st_ld_hazards = False

            self._max_displacement = 1.0

            self.maximize_register_lifetimes = False
            self.minimize_spills = False
            self.move_stalls_to_top = False
            self.move_stalls_to_bottom = False
            self.minimize_register_usage = None
            self.minimize_use_of_extra_registers = None
            self.allow_extra_registers = {}

            self._stalls_allowed = 0
            self._stalls_maximum_attempt = 512
            self._stalls_minimum_attempt = 0
            self._stalls_precision = 0
            self._stalls_timeout_below_precision = None
            self._stalls_first_attempt = 0

            self._model_latencies = True
            self._model_functional_units = True
            self._allow_reordering = True
            self._allow_renaming = True
            self._allow_spills = False
            self._spill_stack_loc_prefix = "STACK_LOC"
            self._spill_type = None

            self.lock()

        @max_displacement.setter
        def max_displacement(self, val):
            self._max_displacement = val

        @stalls_allowed.setter
        def stalls_allowed(self, val):
            self._stalls_allowed = val

        @stalls_maximum_attempt.setter
        def stalls_maximum_attempt(self, val):
            self._stalls_maximum_attempt = val

        @stalls_minimum_attempt.setter
        def stalls_minimum_attempt(self, val):
            self._stalls_minimum_attempt = val

        @stalls_first_attempt.setter
        def stalls_first_attempt(self, val):
            self._stalls_first_attempt = val

        @stalls_precision.setter
        def stalls_precision(self, val):
            self._stalls_precision = val

        @stalls_timeout_below_precision.setter
        def stalls_timeout_below_precision(self, val):
            self._stalls_timeout_below_precision = val

        @model_latencies.setter
        def model_latencies(self, val):
            self._model_latencies = val

        @model_functional_units.setter
        def model_functional_units(self, val):
            self._model_functional_units = val

        @allow_reordering.setter
        def allow_reordering(self, val):
            self._allow_reordering = val

        @allow_renaming.setter
        def allow_renaming(self, val):
            self._allow_renaming = val

        @allow_spills.setter
        def allow_spills(self, val):
            self._allow_spills = val

        @spill_stack_loc_prefix.setter
        def spill_stack_loc_prefix(self, val):
            self._spill_stack_loc_prefix = val

        @spill_type.setter
        def spill_type(self, val):
            self._spill_type = val

        @minimize_spills.setter
        def minimize_spills(self, val):
            self._minimize_spills = val

        @functional_only.setter
        def functional_only(self, val):
            self._model_latencies = val is False
            self._model_functional_units = val is False

    class Hints(NestedPrint, LockAttributes):
        """Subconfiguration for solver hints"""

        @property
        def all_core(self):
            """When SW pipelining is used, hint that all instructions
            should be 'core' instructions (not early/late)."""
            return self._all_core

        @property
        def order_hint_orig_order(self):
            """Hint at using the initial program order for the
            program order variables."""
            return self._order_hint_orig_order

        @property
        def rename_hint_orig_rename(self):
            """Hint at using the initial program order for the
            program order variables."""
            return self._rename_hint_orig_rename

        @property
        def ext_bsearch_remember_successes(self):
            """When using an external binary search, hint previous successful
            optimization.

            See also Config.variable_size."""
            return self._ext_bsearch_remember_successes

        def __init__(self):
            super().__init__()

            self._all_core = True
            self._order_hint_orig_order = False
            self._rename_hint_orig_rename = False
            self._ext_bsearch_remember_successes = False

            self.lock()

        @all_core.setter
        def all_core(self, val):
            self._all_core = val

        @rename_hint_orig_rename.setter
        def rename_hint_orig_rename(self, val):
            self._rename_hint_orig_rename = val

        @order_hint_orig_order.setter
        def order_hint_orig_order(self, val):
            self._order_hint_orig_order = val

    def __init__(self, Arch, Target, logger):
        super().__init__()

        self._arch = Arch
        self._target = Target
        self.logger = logger

        self._sw_pipelining = Config.SoftwarePipelining()
        self._constraints = Config.Constraints()
        self._hints = Config.Hints()

        self._variable_size = False

        self._register_aliases = {}
        self._outputs = set()

        self._inputs_are_outputs = False
        self._rename_inputs = {"arch": "static", "symbolic": "any"}
        self._rename_outputs = {"arch": "static", "symbolic": "any"}

        self._locked_registers = []
        self._reserved_regs = None
        self._reserved_regs_are_locked = True

        self._selftest = True
        self._selftest_iterations = 10
        self._selftest_address_registers = None
        self._selftest_default_memory_size = 1024
        self._selfcheck = True
        self._selfcheck_failure_logfile = None
        self._allow_useless_instructions = False

        # TODO: This should be False by default, but this is a breaking
        # change that requires a lot of examples (where it _is_ safe to
        # apply address offset fixup) to be changed.
        self._unsafe_address_offset_fixup = True

        self._absorb_spills = True

        self._split_heuristic = False
        self._split_heuristic_region = [0.0, 1.0]
        self._split_heuristic_chunks = False
        self._split_heuristic_optimize_seam = 0
        self._split_heuristic_bottom_to_top = False
        self._split_heuristic_factor = 2
        self._split_heuristic_abort_cycle_at_high = None
        self._split_heuristic_abort_cycle_at_low = None
        self._split_heuristic_stepsize = None
        self._split_heuristic_repeat = 1
        self._split_heuristic_preprocess_naive_interleaving = False
        self._split_heuristic_preprocess_naive_interleaving_by_latency = False
        self._split_heuristic_preprocess_naive_interleaving_strategy = "depth"
        self._split_heuristic_estimate_performance = True

        self._compiler_binary = "gcc"
        self._compiler_include_paths = None

        self.keep_tags = True
        self.inherit_macro_comments = False
        self.ignore_tags = False

        self._do_address_fixup = True

        self._with_preprocessor = False
        self._llvm_mca_full = False
        self._llvm_mca_issue_width_overwrite = False
        self._with_llvm_mca_before = False
        self._with_llvm_mca_after = False
        self._max_solutions = 64
        self._timeout = None
        self._retry_timeout = None
        self._ignore_objective = False
        self._objective_precision = 0
        self._objective_lower_bound = None

        # Visualization
        self.indentation = 8
        self.visualize_reordering = True
        self.visualize_expected_performance = True
        self.visualize_show_old_code = False

        self.placeholder_char = "."
        self.early_char = "e"
        self.late_char = "l"
        self.core_char = "*"

        self.mirror_char = "~"

        self.solver_random_seed = 42

        # TODO: Document log_dir and log_model
        self.log_model = None
        self.log_model_only_on_success = True
        self.log_model_dir = "models"

        self.log_model_log_results = True
        self.log_model_results_file = "results.txt"

        self.lock()

    @arch.setter
    def arch(self, val):
        self._arch = val

    @target.setter
    def target(self, val):
        self._target = val

    @sw_pipelining.setter
    def sw_pipelining(self, val):
        self._sw_pipelining = val

    @constraints.setter
    def constraints(self, val):
        self._constraints = val

    @register_aliases.setter
    def register_aliases(self, val):
        self._register_aliases = val

    @outputs.setter
    def outputs(self, val):
        self._outputs = val

    @inputs_are_outputs.setter
    def inputs_are_outputs(self, val):
        self._inputs_are_outputs = val

    @rename_inputs.setter
    def rename_inputs(self, val):
        self._rename_inputs = val
        self._check_rename_config(self._rename_inputs)

    @rename_outputs.setter
    def rename_outputs(self, val):
        self._rename_outputs = val
        self._check_rename_config(self._rename_outputs)

    def _check_rename_config(self, lst):
        assert isinstance(lst, dict)

    @reserved_regs.setter
    def reserved_regs(self, val):
        self._reserved_regs = val

    @reserved_regs_are_locked.setter
    def reserved_regs_are_locked(self, val):
        self._reserved_regs_are_locked = val

    @variable_size.setter
    def variable_size(self, val):
        self._variable_size = val

    @selftest.setter
    def selftest(self, val):
        self._selftest = val

    @selftest_iterations.setter
    def selftest_iterations(self, val):
        self._selftest_iterations = val

    @selftest_address_registers.setter
    def selftest_address_registers(self, val):
        self._selftest_address_registers = val

    @selftest_default_memory_size.setter
    def selftest_default_memory_size(self, val):
        self._selftest_default_memory_size = val

    @selfcheck.setter
    def selfcheck(self, val):
        self._selfcheck = val

    @selfcheck_failure_logfile.setter
    def selfcheck_failure_logfile(self, val):
        self._selfcheck_failure_logfile = val

    @allow_useless_instructions.setter
    def allow_useless_instructions(self, val):
        self._allow_useless_instructions = val

    @unsafe_address_offset_fixup.setter
    def unsafe_address_offset_fixup(self, val):
        self._unsafe_address_offset_fixup = val

    @locked_registers.setter
    def locked_registers(self, val):
        self._locked_registers = val

    @max_solutions.setter
    def max_solutions(self, val):
        self._max_solutions = val

    @with_preprocessor.setter
    def with_preprocessor(self, val):
        self._with_preprocessor = val

    @llvm_mca_issue_width_overwrite.setter
    def llvm_mca_issue_width_overwrite(self, val):
        self._llvm_mca_issue_width_overwrite = val

    @llvm_mca_full.setter
    def llvm_mca_full(self, val):
        self._llvm_mca_full = val

    @with_llvm_mca.setter
    def with_llvm_mca(self, val):
        self._with_llvm_mca_before = val
        self._with_llvm_mca_after = val

    @with_llvm_mca_after.setter
    def with_llvm_mca_after(self, val):
        self._with_llvm_mca_after = val

    @with_llvm_mca_before.setter
    def with_llvm_mca_before(self, val):
        self._with_llvm_mca_before = val

    @compiler_binary.setter
    def compiler_binary(self, val):
        self._compiler_binary = val

    @compiler_include_paths.setter
    def compiler_include_paths(self, val):
        self._compiler_include_paths = val

    @timeout.setter
    def timeout(self, val):
        self._timeout = val

    @retry_timeout.setter
    def retry_timeout(self, val):
        self._retry_timeout = val

    @keep_tags.setter
    def keep_tags(self, val):
        self._keep_tags = val

    @inherit_macro_comments.setter
    def inherit_macro_comments(self, val):
        self._inherit_macro_comments = val

    @ignore_tags.setter
    def ignore_tags(self, val):
        self._ignore_tags = val

    @do_address_fixup.setter
    def do_address_fixup(self, val):
        self._do_address_fixup = val

    @ignore_objective.setter
    def ignore_objective(self, val):
        self._ignore_objective = val

    @objective_precision.setter
    def objective_precision(self, val):
        self._objective_precision = val

    @objective_lower_bound.setter
    def objective_lower_bound(self, val):
        self._objective_lower_bound = val

    @absorb_spills.setter
    def absorb_spills(self, val):
        self._absorb_spills = val

    @split_heuristic.setter
    def split_heuristic(self, val):
        self._split_heuristic = val

    @split_heuristic_factor.setter
    def split_heuristic_factor(self, val):
        self._split_heuristic_factor = float(val)

    @split_heuristic_abort_cycle_at_high.setter
    def split_heuristic_abort_cycle_at_high(self, val):
        self._split_heuristic_abort_cycle_at_high = val

    @split_heuristic_abort_cycle_at_low.setter
    def split_heuristic_abort_cycle_at_low(self, val):
        self._split_heuristic_abort_cycle_at_low = val

    @split_heuristic_stepsize.setter
    def split_heuristic_stepsize(self, val):
        self._split_heuristic_stepsize = float(val)

    @split_heuristic_chunks.setter
    def split_heuristic_chunks(self, val):
        self._split_heuristic_chunks = val

    @split_heuristic_optimize_seam.setter
    def split_heuristic_optimize_seam(self, val):
        self._split_heuristic_optimize_seam = val

    @split_heuristic_bottom_to_top.setter
    def split_heuristic_bottom_to_top(self, val):
        self._split_heuristic_bottom_to_top = val

    @split_heuristic_region.setter
    def split_heuristic_region(self, val):
        self._split_heuristic_region = val

    @split_heuristic_preprocess_naive_interleaving.setter
    def split_heuristic_preprocess_naive_interleaving(self, val):
        self._split_heuristic_preprocess_naive_interleaving = val

    @split_heuristic_preprocess_naive_interleaving_by_latency.setter
    def split_heuristic_preprocess_naive_interleaving_by_latency(self, val):
        self._split_heuristic_preprocess_naive_interleaving_by_latency = val

    @split_heuristic_preprocess_naive_interleaving_strategy.setter
    def split_heuristic_preprocess_naive_interleaving_strategy(self, val):
        self._split_heuristic_preprocess_naive_interleaving_strategy = val

    @split_heuristic_estimate_performance.setter
    def split_heuristic_estimate_performance(self, val):
        self._split_heuristic_estimate_performance = val

    @split_heuristic_repeat.setter
    def split_heuristic_repeat(self, val):
        self._split_heuristic_repeat = val



import logging
import re
import math
import itertools
from enum import Enum
from functools import cache

from unicorn import (
    UC_ARCH_ARM,
    UC_MODE_THUMB,
    UC_MODE_MCLASS,
)

from unicorn.arm_const import (
    UC_ARM_REG_LR,
    UC_ARM_REG_PC,
    UC_ARM_REG_SP,
    UC_ARM_REG_R0,
    UC_ARM_REG_R1,
    UC_ARM_REG_R2,
    UC_ARM_REG_R3,
    UC_ARM_REG_R4,
    UC_ARM_REG_R5,
    UC_ARM_REG_R6,
    UC_ARM_REG_R7,
    UC_ARM_REG_R8,
    UC_ARM_REG_R9,
    UC_ARM_REG_R10,
    UC_ARM_REG_R11,
    UC_ARM_REG_R12,
    UC_ARM_REG_S0,
    UC_ARM_REG_S1,
    UC_ARM_REG_S2,
    UC_ARM_REG_S3,
    UC_ARM_REG_S4,
    UC_ARM_REG_S5,
    UC_ARM_REG_S6,
    UC_ARM_REG_S7,
    UC_ARM_REG_S8,
    UC_ARM_REG_S9,
    UC_ARM_REG_S10,
    UC_ARM_REG_S11,
    UC_ARM_REG_S12,
    UC_ARM_REG_S13,
    UC_ARM_REG_S14,
    UC_ARM_REG_S15,
    UC_ARM_REG_S16,
    UC_ARM_REG_S17,
    UC_ARM_REG_S18,
    UC_ARM_REG_S19,
    UC_ARM_REG_S20,
    UC_ARM_REG_S21,
    UC_ARM_REG_S22,
    UC_ARM_REG_S23,
    UC_ARM_REG_S24,
    UC_ARM_REG_S25,
    UC_ARM_REG_S26,
    UC_ARM_REG_S27,
    UC_ARM_REG_S28,
    UC_ARM_REG_S29,
    UC_ARM_REG_S30,
    UC_ARM_REG_S31,
)

from slothy.helper import SourceLine, Loop
from sympy import simplify

arch_name = "Arm_v7M"
llvm_mca_arch = "arm"
llvm_mc_arch = "thumb"
llvm_mc_attr = "armv7e-m,thumb2,dsp,fpregs"

unicorn_arch = UC_ARCH_ARM
unicorn_mode = UC_MODE_THUMB | UC_MODE_MCLASS


class RegisterType(Enum):
    GPR = 1
    FPR = 2
    FLAGS = 3
    HINT = 4

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.name

    @cache
    def _spillable(reg_type):
        return reg_type in [RegisterType.GPR]

    # TODO: remove workaround (needed for Python 3.9)
    spillable = staticmethod(_spillable)

    @staticmethod
    def callee_saved_registers():
        return [f"r{i}" for i in range(4, 12)] + [f"s{i}" for i in range(0, 16)]

    @staticmethod
    def unicorn_link_register():
        return UC_ARM_REG_LR

    @staticmethod
    def unicorn_program_counter():
        return UC_ARM_REG_PC

    @staticmethod
    def unicorn_stack_pointer():
        return UC_ARM_REG_SP

    @cache
    def _unicorn_reg_by_name(reg):
        """Converts string name of register into numerical identifiers used
        within the unicorn engine"""

        d = {
            "r0": UC_ARM_REG_R0,
            "r1": UC_ARM_REG_R1,
            "r2": UC_ARM_REG_R2,
            "r3": UC_ARM_REG_R3,
            "r4": UC_ARM_REG_R4,
            "r5": UC_ARM_REG_R5,
            "r6": UC_ARM_REG_R6,
            "r7": UC_ARM_REG_R7,
            "r8": UC_ARM_REG_R8,
            "r9": UC_ARM_REG_R9,
            "r10": UC_ARM_REG_R10,
            "r11": UC_ARM_REG_R11,
            "r12": UC_ARM_REG_R12,
            "r13": UC_ARM_REG_SP,
            "r14": UC_ARM_REG_LR,
            "s0": UC_ARM_REG_S0,
            "s1": UC_ARM_REG_S1,
            "s2": UC_ARM_REG_S2,
            "s3": UC_ARM_REG_S3,
            "s4": UC_ARM_REG_S4,
            "s5": UC_ARM_REG_S5,
            "s6": UC_ARM_REG_S6,
            "s7": UC_ARM_REG_S7,
            "s8": UC_ARM_REG_S8,
            "s9": UC_ARM_REG_S9,
            "s10": UC_ARM_REG_S10,
            "s11": UC_ARM_REG_S11,
            "s12": UC_ARM_REG_S12,
            "s13": UC_ARM_REG_S13,
            "s14": UC_ARM_REG_S14,
            "s15": UC_ARM_REG_S15,
            "s16": UC_ARM_REG_S16,
            "s17": UC_ARM_REG_S17,
            "s18": UC_ARM_REG_S18,
            "s19": UC_ARM_REG_S19,
            "s20": UC_ARM_REG_S20,
            "s21": UC_ARM_REG_S21,
            "s22": UC_ARM_REG_S22,
            "s23": UC_ARM_REG_S23,
            "s24": UC_ARM_REG_S24,
            "s25": UC_ARM_REG_S25,
            "s26": UC_ARM_REG_S26,
            "s27": UC_ARM_REG_S27,
            "s28": UC_ARM_REG_S28,
            "s29": UC_ARM_REG_S29,
            "s30": UC_ARM_REG_S30,
            "s31": UC_ARM_REG_S31,
        }
        return d.get(reg, None)

    # TODO: remove workaround (needed for Python 3.9)
    unicorn_reg_by_name = staticmethod(_unicorn_reg_by_name)

    @cache
    def _list_registers(
        reg_type, only_extra=False, only_normal=False, with_variants=False
    ):
        """Return the list of all registers of a given type"""

        gprs_normal = [f"r{i}" for i in range(15)]
        fprs_normal = [f"s{i}" for i in range(32)]

        gprs_extra = []
        fprs_extra = []

        gprs = []
        fprs = []
        # TODO: What are hints?
        hints = (
            [f"t{i}" for i in range(100)]
            + [f"t{i}{j}" for i in range(8) for j in range(8)]
            + [f"t{i}_{j}" for i in range(16) for j in range(16)]
        )

        flags = ["flags"]
        if not only_extra:
            gprs += gprs_normal
            fprs += fprs_normal
        if not only_normal:
            gprs += gprs_extra
            fprs += fprs_extra

        return {
            RegisterType.GPR: gprs,
            RegisterType.FPR: fprs,
            RegisterType.HINT: hints,
            RegisterType.FLAGS: flags,
        }[reg_type]

    # TODO: remove workaround (needed for Python 3.9)
    list_registers = staticmethod(_list_registers)

    @staticmethod
    def find_type(r):
        """Find type of architectural register"""

        if r.startswith("hint_"):
            return RegisterType.HINT

        for ty in RegisterType:
            if r in RegisterType.list_registers(ty):
                return ty

        return None

    @staticmethod
    def is_renamed(ty):
        """Indicate if register type should be subject to renaming"""
        if ty == RegisterType.HINT:
            return False
        return True

    @staticmethod
    def from_string(string):
        """Find registe type from string"""
        string = string.lower()
        return {
            "fpr": RegisterType.FPR,
            "gpr": RegisterType.GPR,
            "hint": RegisterType.HINT,
            "flags": RegisterType.FLAGS,
        }.get(string, None)

    @staticmethod
    def default_reserved():
        """Return the list of registers that should be reserved by default"""
        # r13 is the stack pointer
        return set(["flags", "r13"] + RegisterType.list_registers(RegisterType.HINT))

    @staticmethod
    def default_aliases():
        "Register aliases used by the architecture"
        return {
            "lr": "r14",
        }


# TODO: Comparison can also be done with {add,sub,...}s
class Branch:
    """Helper for emitting branches"""

    @staticmethod
    def if_equal(cnt, val, lbl):
        """Emit assembly for a branch-if-equal sequence"""
        yield f"cmp {cnt}, #{val}"
        yield f"beq {lbl}"

    @staticmethod
    def if_greater_equal(cnt, val, lbl):
        """Emit assembly for a branch-if-greater-equal sequence"""
        yield f"cmp {cnt}, #{val}"
        yield f"bge {lbl}"

    @staticmethod
    def unconditional(lbl):
        """Emit unconditional branch"""
        yield f"b {lbl}"


class VmovCmpLoop(Loop):
    """
    Loop ending in a vmov, a compare, and a branch.

    The modification to the value we compare against happens inside the loop
    body. The value that is being compared to is stashed to a floating point
    register before the loop starts and therefore needs to be recovered before
    the comparison.

    .. warning::

        This type of loop is experimental as slothy has no knowledge about
        what happens inside the loop boundary! Especially, a register is written
        inside the boundary which may be used for renaming by slothy. Use with
        caution.

    Example:

    .. code-block:: asm

        loop_lbl:
           {code}
           vmov <end>, <endf>
           cmp <cnt>, <end>
           (cbnz|bnz|bne) loop_lbl

    where cnt is the loop counter in lr.
    """

    def __init__(self, lbl="lbl", lbl_start="1", lbl_end="2", loop_init="lr") -> None:
        super().__init__(lbl_start=lbl_start, lbl_end=lbl_end, loop_init=loop_init)
        self.lbl = lbl
        self.lbl_regex = r"^\s*(?P<label>[\w\.]+)\s*:(?P<remainder>.*)$"
        self.end_regex = (
            r"^\s*vmov(?:\.w)?\s+(?P<end>\w+),\s*(?P<endf>\w+)",
            r"^\s*cmp(?:\.w)?\s+(?P<cnt>\w+),\s*(?P<end>\w+)",
            rf"^\s*(cbnz|cbz|bne)(?:\.w)?\s+{lbl}",
        )

    def start(
        self,
        loop_cnt,
        indentation=0,
        fixup=0,
        unroll=1,
        jump_if_empty=None,
        preamble_code=None,
        body_code=None,
        postamble_code=None,
        register_aliases=None,
    ):
        """Emit starting instruction(s) and jump label for loop"""
        indent = " " * indentation
        if unroll > 1:
            assert unroll in [1, 2, 4, 8, 16, 32]
            yield (
                f"{indent}lsr {self.additional_data['end']}, "
                f"{self.additional_data['end']}, #{int(math.log2(unroll))}"
            )

        # Find out by how much the loop counter is modified in one iteration
        inc_per_iter = 0
        if body_code is not None:
            try:
                loop_cnt_reg = register_aliases[loop_cnt]
            except KeyError:
                loop_cnt_reg = loop_cnt
            for line in body_code:
                if line.text == "":
                    continue
                inst = Instruction.parser(line)
                if (
                    loop_cnt_reg.lower() == inst[0].addr
                    and inst[0].increment is not None
                ):
                    inc_per_iter = inc_per_iter + simplify(inst[0].increment)
        logging.debug(
            f"Loop counter {loop_cnt} is incremented by {inc_per_iter} per iteration"
        )
        # Check whether instructions modifying the loop count moved to
        # pre/postamble and adjust the fixup based on that.
        # new_fixup = 0
        # if postamble_code is not None:
        #     new_fixup = 0
        #     for l in postamble_code:
        #         if l.text == "":
        #             continue
        #         inst = Instruction.parser(l)
        #         if loop_cnt in inst[0].args_in_out and inst[0].increment is not None:
        #             new_fixup = new_fixup + simplify(inst[0].increment)

        # if new_fixup != 0 or fixup != 0:
        if fixup != 0:
            yield f"{indent}push {{{self.additional_data['end']}}}"
            yield (
                f"{indent}vmov {self.additional_data['end']}, "
                f"{self.additional_data['endf']}"
            )

        # if new_fixup != 0:
        #     yield f"{indent}sub {self.additional_data['end']},
        #           {self.additional_data['end']}, #{new_fixup}"
        if fixup != 0:
            yield (
                f"{indent}sub {self.additional_data['end']}, "
                f"{self.additional_data['end']}, #{fixup*inc_per_iter}"
            )
        # if new_fixup != 0 or fixup != 0:
        if fixup != 0:
            yield (
                f"{indent}vmov {self.additional_data['endf']}, "
                f"{self.additional_data['end']}"
            )
            yield f"{indent}pop {{{self.additional_data['end']}}}"
        if jump_if_empty is not None:
            yield f"cbz {loop_cnt}, {jump_if_empty}"
        yield f"{self.lbl}:"

    def end(self, other, indentation=0):
        """Emit compare-and-branch at the end of the loop"""
        indent = " " * indentation
        lbl_start = self.lbl
        if lbl_start.isdigit():
            lbl_start += "b"

        yield f'{indent}vmov {other["end"]}, {other["endf"]}'
        yield f'{indent}cmp {other["cnt"]}, {other["end"]}'
        yield f"{indent}bne {lbl_start}"


class BranchLoop(Loop):
    """
    More general loop type that just considers the branch instruction as part
    of the boundary.
    This can help to improve performance as the instructions that belong to
    handling the loop can be considered by SLOTHY as well.

    .. note::

        This loop type is still rather experimental. It has a lot of logics
        inside as it needs to be able to "understand" a variety of different
        ways to express loops, e.g., how counters get incremented, how
        registers marking the end of the
        loop need to be modified in case of software pipelining etc.
        Currently, this type covers the three other types we offer
        above, namely `SubsLoop`, `CmpLoop`, and `VmovCmpLoop`.

    For examples, we refer to the classes `SubsLoop`, `CmpLoop`, and
    `VmovCmpLoop`.
    """

    def __init__(self, lbl="lbl", lbl_start="1", lbl_end="2", loop_init="lr") -> None:
        super().__init__(lbl_start=lbl_start, lbl_end=lbl_end, loop_init=loop_init)
        self.lbl = lbl
        self.lbl_regex = r"^\s*(?P<label>[\w\.]+)\s*:(?P<remainder>.*)$"
        # Defines the end of the loop, boolean indicates whether the instruction
        # shall be considered part of the body or not.
        self.end_regex = ((rf"^\s*(cbnz|cbz|bne)(?:\.w)?\s+{lbl}", True),)

    def start(
        self,
        loop_cnt,
        indentation=0,
        fixup=0,
        unroll=1,
        jump_if_empty=None,
        preamble_code=None,
        body_code=None,
        postamble_code=None,
        register_aliases=None,
    ):
        """Emit starting instruction(s) and jump label for loop"""
        indent = " " * indentation
        if body_code is None:
            logging.debug("No body code in loop start: Just printing label.")
            yield f"{self.lbl}:"
            return
        # Identify the register that is used as a loop counter
        body_code = [line for line in body_code if line.text != ""]
        for line in body_code:
            inst = Instruction.parser(line)
            # Flags are set through cmp
            # LIMITATION: By convention, we require the first argument to be the
            # "counter" and the second the one marking the iteration end.
            if isinstance(inst[0], cmp):
                # Assume this mapping
                loop_cnt_reg = inst[0].args_in[0]
                loop_end_reg = inst[0].args_in[1]
                logging.debug(
                    f"Assuming {loop_cnt_reg} as counter register and {loop_end_reg} "
                    "as end register."
                )
                break
            # Flags are set through subs
            elif isinstance(inst[0], subs_imm_short):
                loop_cnt_reg = inst[0].args_in_out[0]
                loop_end_reg = inst[0].args_in_out[0]
                break
            elif isinstance(inst[0], subs_imm):
                loop_cnt_reg = inst[0].args_out[0]
                loop_end_reg = inst[0].args_out[0]
                break

        # Find FPR that is used to stash the loop end incase it's vmov loop
        loop_end_reg_fpr = None
        for li, l in enumerate(body_code):
            inst = Instruction.parser(l)
            # Flags are set through cmp
            if isinstance(inst[0], vmov_gpr):
                if loop_end_reg in inst[0].args_out:
                    logging.debug(f"Copying from {inst[0].args_in} to {loop_end_reg}")
                    loop_end_reg_fpr = inst[0].args_in[0]

            # The last vmov occurance before the cmp that writes to the register
            # we compare to will be the right one. The same GPR could be written
            # previously due to renaming, before it becomes the value used in
            # the cmp.
            if isinstance(inst[0], cmp):
                break

        if unroll > 1:
            assert unroll in [1, 2, 4, 8, 16, 32]
            yield f"{indent}lsr {loop_end_reg}, {loop_end_reg}, #{int(math.log2(unroll))}"

        inc_per_iter = 0
        for line in body_code:
            inst = Instruction.parser(line)
            # Increment happens through pointer modification
            if loop_cnt_reg.lower() == inst[0].addr and inst[0].increment is not None:
                inc_per_iter = inc_per_iter + simplify(inst[0].increment)
            # Increment through explicit modification
            elif (
                loop_cnt_reg.lower() in (inst[0].args_out + inst[0].args_in_out)
                and inst[0].immediate is not None
            ):
                # TODO: subtract if we have a subtraction
                inc_per_iter = inc_per_iter + simplify(inst[0].immediate)
        logging.debug(
            f"Loop counter {loop_cnt_reg} is incremented by {inc_per_iter} per iteration."
        )

        if fixup != 0 and loop_end_reg_fpr is not None:
            yield f"{indent}push {{{loop_end_reg}}}"
            yield f"{indent}vmov {loop_end_reg}, {loop_end_reg_fpr}"

        if fixup != 0:
            yield f"{indent}sub {loop_end_reg}, {loop_end_reg}, #{fixup*inc_per_iter}"

        if fixup != 0 and loop_end_reg_fpr is not None:
            yield f"{indent}vmov {loop_end_reg_fpr}, {loop_end_reg}"
            yield f"{indent}pop {{{loop_end_reg}}}"

        if jump_if_empty is not None:
            yield f"cbz {loop_cnt}, {jump_if_empty}"
        yield f"{self.lbl}:"

    def end(self, other, indentation=0):
        """Nothing to do here"""
        yield ""


class CmpLoop(Loop):
    """
    Loop ending in a compare and a branch.
    The modification to the value we compare against happens inside the loop body.
    WARNING: This type of loop is experimental as slothy has no knowledge about
    what happens inside the loop boundary! Use with caution.

    Example:

    .. code-block:: asm

        loop_lbl:
           {code}
           cmp <cnt>, <end>
           (cbnz|bnz|bne) loop_lbl

    where cnt is the loop counter in lr.
    """

    def __init__(self, lbl="lbl", lbl_start="1", lbl_end="2", loop_init="lr") -> None:
        super().__init__(lbl_start=lbl_start, lbl_end=lbl_end, loop_init=loop_init)
        self.lbl_regex = r"^\s*(?P<label>[\w\.]+)\s*:(?P<remainder>.*)$"
        self.end_regex = (
            r"^\s*cmp(?:\.w)?\s+(?P<cnt>\w+),\s*(?P<end>\w+)",
            rf"^\s*(cbnz|cbz|bne)(?:\.w)?\s+{lbl}",
        )

    def start(
        self,
        loop_cnt,
        indentation=0,
        fixup=0,
        unroll=1,
        jump_if_empty=None,
        preamble_code=None,
        body_code=None,
        postamble_code=None,
        register_aliases=None,
    ):
        """Emit starting instruction(s) and jump label for loop"""
        indent = " " * indentation
        if unroll > 1:
            assert unroll in [1, 2, 4, 8, 16, 32]
            yield (
                f"{indent}lsr {self.additional_data['end']}, "
                f"{self.additional_data['end']}, #{int(math.log2(unroll))}"
            )

        # Find out by how much the loop counter is modified in one iteration
        inc_per_iter = 0
        if body_code is not None:
            try:
                loop_cnt_reg = register_aliases[loop_cnt]
            except KeyError:
                loop_cnt_reg = loop_cnt
            for line in body_code:
                if line.text == "":
                    continue
                inst = Instruction.parser(line)
                if (
                    loop_cnt_reg.lower() == inst[0].addr
                    and inst[0].increment is not None
                ):
                    inc_per_iter = inc_per_iter + simplify(inst[0].increment)
        logging.debug(
            f"Loop counter {loop_cnt} is incremented by {inc_per_iter} per iteration"
        )

        # Check whether instructions modifying the loop count moved to
        # pre/postamble and adjust the fixup based on that.
        # new_fixup = 0
        # if postamble_code is not None:
        #     new_fixup = 0
        #     for l in postamble_code:
        #         if l.text == "":
        #             continue
        #         inst = Instruction.parser(l)
        #         if loop_cnt in inst[0].args_in_out and inst[0].increment is not None:
        #             new_fixup = new_fixup + simplify(inst[0].increment)

        # if new_fixup != 0:
        #     yield f"{indent}sub {self.additional_data['end']},
        #  {self.additional_data['end']}, #{new_fixup}"

        if fixup != 0:
            yield (
                f"{indent}sub {self.additional_data['end']}, "
                f"{self.additional_data['end']}, #{fixup*inc_per_iter}"
            )

        if jump_if_empty is not None:
            yield f"cbz {loop_cnt}, {jump_if_empty}"
        yield f"{self.lbl_start}:"

    def end(self, other, indentation=0):
        """Emit compare-and-branch at the end of the loop"""
        indent = " " * indentation
        lbl_start = self.lbl_start
        if lbl_start.isdigit():
            lbl_start += "b"

        yield f'{indent}cmp {other["cnt"]}, {other["end"]}'
        yield f"{indent}bne {lbl_start}"


class SubsLoop(Loop):
    """
    Loop ending in a flag setting subtraction and a branch.

    Example:

    .. code-block::

        loop_lbl:
           {code}
           sub[s] <cnt>, <cnt>, #1
           (cbnz|bnz|bne) loop_lbl

    where cnt is the loop counter in lr.
    """

    def __init__(self, lbl_start="1", lbl_end="2", loop_init="lr") -> None:
        super().__init__(lbl_start=lbl_start, lbl_end=lbl_end, loop_init=loop_init)
        self.lbl_regex = r"^\s*(?P<label>[\w\.]+)\s*:(?P<remainder>.*)$"
        self.end_regex = (
            r"^\s*sub[s]?(?:\.w)?\s+(?P<cnt>\w+),(?:\s*(?P<reg1>\w+),)?\s*(?P<imm>#1)",
            rf"^\s*(cbnz|cbz|bne)(?:\.w)?\s+{lbl_start}",
        )

    def start(
        self,
        loop_cnt,
        indentation=0,
        fixup=0,
        unroll=1,
        jump_if_empty=None,
        preamble_code=None,
        body_code=None,
        postamble_code=None,
        register_aliases=None,
    ):
        """Emit starting instruction(s) and jump label for loop"""
        indent = " " * indentation
        if unroll > 1:
            assert unroll in [1, 2, 4, 8, 16, 32]
            yield f"{indent}lsr {loop_cnt}, {loop_cnt}, #{int(math.log2(unroll))}"
        if fixup != 0:
            yield f"{indent}sub {loop_cnt}, {loop_cnt}, #{fixup}"
        if jump_if_empty is not None:
            yield f"cbz {loop_cnt}, {jump_if_empty}"
        yield f"{self.lbl_start}:"

    def end(self, other, indentation=0):
        """Emit compare-and-branch at the end of the loop"""
        indent = " " * indentation
        lbl_start = self.lbl_start
        if lbl_start.isdigit():
            lbl_start += "b"
        if other["reg1"] is None:
            yield f'{indent}subs {other["cnt"]}, #1'
        else:
            # `subs` sets flags
            yield f'{indent}subs {other["cnt"]}, {other["reg1"]}, {other["imm"]}'
        yield f"{indent}bne {lbl_start}"


class FatalParsingException(Exception):
    """A fatal error happened during instruction parsing"""


class UnknownInstruction(Exception):
    """The parent instruction class for the given object could not be found"""


class UnknownRegister(Exception):
    """The register could not be found"""


class Instruction:

    class ParsingException(Exception):
        """An attempt to parse an assembly line as a specific instruction failed

        This is a frequently encountered exception since assembly lines are parsed by
        trial and error, iterating over all instruction parsers."""

        def __init__(self, err=None):
            super().__init__(err)

    def __init__(
        self, *, mnemonic, arg_types_in=None, arg_types_in_out=None, arg_types_out=None
    ):

        if arg_types_in is None:
            arg_types_in = []
        if arg_types_out is None:
            arg_types_out = []
        if arg_types_in_out is None:
            arg_types_in_out = []

        self.mnemonic = mnemonic

        self.args_out_combinations = None
        self.args_in_combinations = None
        self.args_in_out_combinations = None
        self.args_in_out_different = None
        self.args_inout_out_different = None
        self.args_in_inout_different = None

        self.arg_types_in = arg_types_in
        self.arg_types_out = arg_types_out
        self.arg_types_in_out = arg_types_in_out
        self.num_in = len(arg_types_in)
        self.num_out = len(arg_types_out)
        self.num_in_out = len(arg_types_in_out)

        self.args_out_restrictions = [None for _ in range(self.num_out)]
        self.args_in_restrictions = [None for _ in range(self.num_in)]
        self.args_in_out_restrictions = [None for _ in range(self.num_in_out)]

        self.args_in = []
        self.args_out = []
        self.args_in_out = []

        self.addr = None
        self.increment = None
        self.pre_index = None
        self.offset_adjustable = True

        self.immediate = None
        self.datatype = None
        self.index = None
        self.flag = None
        self.width = None
        self.barrel = None
        self.label = None
        self.reg_list = None

    def extract_read_writes(self):
        """Extracts 'reads'/'writes' clauses from the source line of the instruction"""

        src_line = self.source_line

        def hint_register_name(tag):
            return f"hint_{tag}"

        # Check if the source line is tagged as reading/writing from memory
        def add_memory_write(tag):
            self.num_out += 1
            self.pattern_outputs.append((hint_register_name(tag), RegisterType.HINT))
            self.args_out_restrictions.append(None)
            self.args_out.append(hint_register_name(tag))
            self.arg_types_out.append(RegisterType.HINT)

        def add_memory_read(tag):
            self.num_in += 1
            self.pattern_inputs.append((hint_register_name(tag), RegisterType.HINT))
            self.args_in_restrictions.append(None)
            self.args_in.append(hint_register_name(tag))
            self.arg_types_in.append(RegisterType.HINT)

        write_tags = src_line.tags.get("writes", [])
        read_tags = src_line.tags.get("reads", [])

        if not isinstance(write_tags, list):
            write_tags = [write_tags]

        if not isinstance(read_tags, list):
            read_tags = [read_tags]

        for w in write_tags:
            add_memory_write(w)

        for r in read_tags:
            add_memory_read(r)

        return self

    def global_parsing_cb(self, a, log=None):
        """Parsing callback triggered after DataFlowGraph parsing which allows
        modification of the instruction in the context of the overall computation.

        This is primarily used to remodel input-outputs as outputs in jointly destructive
        instruction patterns (See Section 4.4, https://eprint.iacr.org/2022/1303.pdf).
        """
        _ = log  # log is not used
        return False

    def global_fusion_cb(self, a, log=None):
        """Fusion callback triggered after DataFlowGraph parsing which allows fusing
        of the instruction in the context of the overall computation.

        This can be used e.g. to detect eor-eor pairs and replace them by eor3."""
        _ = log  # log is not used
        return False

    def write(self):
        """Write the instruction"""
        args = self.args_out + self.args_in_out + self.args_in
        return self.mnemonic + " " + ", ".join(args)

    @staticmethod
    def unfold_abbrevs(mnemonic):
        return mnemonic

    def _is_instance_of(self, inst_list):
        for inst in inst_list:
            if isinstance(self, inst):
                return True
        return False

    def is_ldr(self):
        return self._is_instance_of(
            [
                ldr,
                ldr_with_imm,
                ldr_with_imm_stack,
                ldr_with_postinc,
                ldr_with_inc_writeback,
            ]
        )

    def is_load(self):
        """Indicates if an instruction is a load instruction"""
        return self._is_instance_of(
            [
                ldr,
                ldr_with_imm,
                ldrb_with_imm,
                ldrh_with_imm,
                ldr_with_imm_stack,
                ldr_with_postinc,
                ldrh_with_postinc,
                ldrb_with_postinc,
                ldrd_imm,
                ldrd_with_postinc,
                ldr_with_inc_writeback,
                ldm_interval,
                ldm_interval_inc_writeback,
                vldr_with_imm,
                vldr_with_postinc,
                vldm_interval_inc_writeback,
            ]
        )

    def is_store(self):
        """Indicates if an instruction is a store instruction"""
        return self._is_instance_of(
            [
                str_no_off,
                strh_with_imm,
                str_with_imm,
                str_with_imm_stack,
                str_with_postinc,
                strh_with_postinc,
                stm_interval_inc_writeback,
            ]
        )

    def is_load_store_instruction(self):
        """Indicates if an instruction is a load or store instruction"""
        return self.is_load() or self.is_store()

    @classmethod
    def make(cls, src):
        """Abstract factory method parsing a string into an instruction instance."""

    @staticmethod
    def build(c: any, src: str, mnemonic: str, **kwargs: list) -> "Instruction":
        """Attempt to parse a string as an instance of an instruction.


        :param c: The target instruction the string should be attempted to be parsed as.
        :type c: any
        :param src: The string to parse.
        :type src: str
        :param mnemonic: The mnemonic of instruction c
        :type mnemonic: str
        :param **kwargs: Additional arguments to pass to the constructor of c.
        :type **kwargs: list

        :return: Upon success, the result of parsing src as an instance of c.
        :rtype: Instruction

        :raises Instruction.ParsingException: The str argument cannot be parsed as an
                instance of c.
        :raises FatalParsingException: A fatal error during parsing happened
                that's likely a bug in the model.
        """

        if src.split(" ")[0] != mnemonic:
            raise Instruction.ParsingException(
                f"Mnemonic does not match: {src.split(' ')[0]} vs. {mnemonic}"
            )

        obj = c(mnemonic=mnemonic, **kwargs)

        # Replace <dt> by list of all possible datatypes
        mnemonic = Instruction.unfold_abbrevs(obj.mnemonic)

        expected_args = obj.num_in + obj.num_out + obj.num_in_out
        regexp_txt = rf"^\s*{mnemonic}"
        if expected_args > 0:
            regexp_txt += r"\s+"
        regexp_txt += ",".join([r"\s*(\w+)\s*" for _ in range(expected_args)])
        regexp = re.compile(regexp_txt)

        p = regexp.match(src)
        if p is None:
            raise Instruction.ParsingException(
                f"Doesn't match basic instruction template {regexp_txt}"
            )

        operands = list(p.groups())

        if obj.num_out > 0:
            obj.args_out = operands[: obj.num_out]
            idx_args_in = obj.num_out
        elif obj.num_in_out > 0:
            obj.args_in_out = operands[: obj.num_in_out]
            idx_args_in = obj.num_in_out
        else:
            idx_args_in = 0

        obj.args_in = operands[idx_args_in:]

        if not len(obj.args_in) == obj.num_in:
            raise FatalParsingException(
                f"Something wrong parsing {src}: Expect {obj.num_in} input,"
                f" but got {len(obj.args_in)} ({obj.args_in})"
            )

        return obj

    @staticmethod
    def parser(src_line):
        """Global factory method parsing an assembly line into an instance
        of a subclass of Instruction."""
        insts = []
        exceptions = {}
        instnames = []

        src = src_line.text.strip()

        # Iterate through all derived classes and call their parser
        # until one of them hopefully succeeds
        for inst_class in Instruction.all_subclass_leaves:
            try:
                inst = inst_class.make(src)
                instnames = [inst_class.__name__]
                insts = [inst]
                break
            except Instruction.ParsingException as e:
                exceptions[inst_class.__name__] = e

        for i in insts:
            i.source_line = src_line

            # Mark as branch for BranchLoop
            if isinstance(i, Armv7mBranch):
                i.source_line.tags["branch"] = True

            i.extract_read_writes()

        if len(insts) == 0:
            logging.error("Failed to parse instruction %s", src)
            logging.error("A list of attempted parsers and their exceptions follows.")
            for i, e in exceptions.items():
                msg = f"* {i + ':':20s} {e}"
                logging.error(msg)
            raise Instruction.ParsingException(
                f"Couldn't parse {src}\nYou may need to add support "
                "for a new instruction (variant)?"
            )

        logging.debug("Parsing result for '%s': %s", src, instnames)
        return insts

    def __repr__(self):
        return self.write()


class Armv7mInstruction(Instruction):
    """Abstract class representing Armv7m instructions"""

    PARSERS = {}

    @staticmethod
    def _unfold_pattern(src):
        src = re.sub(r", +", ",", src)
        src = re.sub(r"\.", "\\\\s*\\\\.\\\\s*", src)
        src = re.sub(r"\[", "\\\\s*\\\\[\\\\s*", src)
        src = re.sub(r"\]", "\\\\s*\\\\]\\\\s*", src)

        def pattern_transform(g):
            return (
                f"([{g.group(1).lower()}{g.group(1)}]"
                f"(?P<raw_{g.group(1)}{g.group(2)}>[0-9_][0-9_]*)|"
                f"([{g.group(1).lower()}{g.group(1)}]"
                f"<(?P<symbol_{g.group(1)}{g.group(2)}>\\w+)>))"
            )

        src = re.sub(r"<([RS])(\w+)>", pattern_transform, src)

        # Replace <key> or <key0>, <key1>, ... with pattern
        def replace_placeholders(src, mnemonic_key, regexp, group_name):
            prefix = f"<{mnemonic_key}"
            pattern = f"<{mnemonic_key}>"

            def pattern_i(i):
                return f"<{mnemonic_key}{i}>"

            cnt = src.count(prefix)
            if cnt > 1:
                for i in range(cnt):
                    src = re.sub(pattern_i(i), f"(?P<{group_name}{i}>{regexp})", src)
            else:
                src = re.sub(pattern, f"(?P<{group_name}>{regexp})", src)

            return src

        flaglist = [
            "eq",
            "ne",
            "cs",
            "hs",
            "cc",
            "lo",
            "mi",
            "pl",
            "vs",
            "vc",
            "hi",
            "ls",
            "ge",
            "lt",
            "gt",
            "le",
        ]

        flag_pattern = "|".join(flaglist)
        # TODO: Notion of dt can be placed with notion for size in FP instructions
        dt_pattern = "(?:|2|4|8|16)(?:B|H|S|D|b|h|s|d)"
        imm_pattern = (
            "(#(\\\\w|\\\\s|/| |-|\\*|\\+|\\(|\\)|=|<<|>>)+)"
            "|"
            "(((0[xb])?[0-9a-fA-F]+|/| |-|\\*|\\+|\\(|\\)|=|<<|>>)+)"
        )
        index_pattern = "[0-9]+"
        width_pattern = r"(?:\.w|\.n|)"
        barrel_pattern = "(?:lsl|ror|lsr|asr)\\\\s*"
        label_pattern = r"(?:\\w+)"

        # reg_list is <range>(,<range>)*
        # range is [rs]NN(-rsMM)?
        range_pat = "([rs]\\\\d+)(-[rs](\\\\d+))?"
        reg_list_pattern = "{" + range_pat + "(," + range_pat + ")*" + "}"

        src = re.sub(" ", "\\\\s+", src)
        src = re.sub(",", "\\\\s*,\\\\s*", src)

        src = replace_placeholders(src, "imm", imm_pattern, "imm")
        src = replace_placeholders(src, "dt", dt_pattern, "datatype")
        src = replace_placeholders(src, "index", index_pattern, "index")
        src = replace_placeholders(
            src, "flag", flag_pattern, "flag"
        )  # TODO: Are any changes required for IT syntax?
        src = replace_placeholders(src, "width", width_pattern, "width")
        src = replace_placeholders(src, "barrel", barrel_pattern, "barrel")
        src = replace_placeholders(src, "label", label_pattern, "label")
        src = replace_placeholders(src, "reg_list", reg_list_pattern, "reg_list")

        src = r"\s*" + src + r"\s*(//.*)?\Z"
        return src

    @staticmethod
    def _build_parser(src):
        regexp_txt = Armv7mInstruction._unfold_pattern(src)
        regexp = re.compile(regexp_txt)

        def _parse(line):
            regexp_result = regexp.match(line)
            if regexp_result is None:
                raise Instruction.ParsingException(
                    f"Does not match instruction pattern {src}" f"[regex: {regexp_txt}]"
                )
            res = regexp.match(line).groupdict()
            items = list(res.items())
            for k, v in items:
                for prefix in ["symbol_", "raw_"]:
                    if k.startswith(prefix):
                        del res[k]
                        if v is None:
                            continue
                        k = k[len(prefix) :]
                        res[k] = v
            return res

        return _parse

    @staticmethod
    def get_parser(pattern):
        """Build parser for given AArch64 instruction pattern"""
        if pattern in Armv7mInstruction.PARSERS:
            return Armv7mInstruction.PARSERS[pattern]
        parser = Armv7mInstruction._build_parser(pattern)
        Armv7mInstruction.PARSERS[pattern] = parser
        return parser

    @cache
    def __infer_register_type(ptrn):
        if ptrn[0].upper() in ["R"]:
            return RegisterType.GPR
        if ptrn[0].upper() in ["S"]:
            return RegisterType.FPR
        if ptrn[0].upper() in ["T"]:
            return RegisterType.HINT
        raise FatalParsingException(f"Unknown pattern: {ptrn}")

    # TODO: remove workaround (needed for Python 3.9)
    _infer_register_type = staticmethod(__infer_register_type)

    def __init__(
        self,
        pattern,
        *,
        inputs=None,
        outputs=None,
        in_outs=None,
        modifiesFlags=False,
        dependsOnFlags=False,
    ):

        self.mnemonic = pattern.split(" ")[0]

        if inputs is None:
            inputs = []
        if outputs is None:
            outputs = []
        if in_outs is None:
            in_outs = []
        arg_types_in = [Armv7mInstruction._infer_register_type(r) for r in inputs]
        arg_types_out = [Armv7mInstruction._infer_register_type(r) for r in outputs]
        arg_types_in_out = [Armv7mInstruction._infer_register_type(r) for r in in_outs]

        if modifiesFlags:
            arg_types_out += [RegisterType.FLAGS]
            outputs += ["flags"]

        if dependsOnFlags:
            arg_types_in += [RegisterType.FLAGS]
            inputs += ["flags"]

        super().__init__(
            mnemonic=pattern,
            arg_types_in=arg_types_in,
            arg_types_out=arg_types_out,
            arg_types_in_out=arg_types_in_out,
        )

        self.inputs = inputs
        self.outputs = outputs
        self.in_outs = in_outs

        self.pattern = pattern
        assert len(inputs) == len(arg_types_in)
        self.pattern_inputs = list(zip(inputs, arg_types_in))
        assert len(outputs) == len(arg_types_out)
        self.pattern_outputs = list(zip(outputs, arg_types_out))
        assert len(in_outs) == len(arg_types_in_out)
        self.pattern_in_outs = list(zip(in_outs, arg_types_in_out))

    @staticmethod
    def _to_reg(ty, s):
        if ty == RegisterType.GPR:
            c = "r"
        elif ty == RegisterType.FPR:
            c = "s"
        elif ty == RegisterType.HINT:
            c = "t"
        else:
            assert False
        if s.replace("_", "").isdigit():
            return f"{c}{s}"
        return s

    @staticmethod
    def _build_pattern_replacement(s, ty, arg):
        if ty == RegisterType.GPR:
            if arg[0] != "r":
                return f"{s[0].upper()}<{arg}>"
            return s[0].lower() + arg[1:]
        if ty == RegisterType.FPR:
            if arg[0] != "s":
                return f"{s[0].upper()}<{arg}>"
            return s[0].lower() + arg[1:]
        if ty == RegisterType.HINT:
            if arg[0] != "t":
                return f"{s[0].upper()}<{arg}>"
            return s[0].lower() + arg[1:]
        raise FatalParsingException(f"Unknown register type ({s}, {ty}, {arg})")

    @staticmethod
    def _instantiate_pattern(s, ty, arg, out):
        if ty == RegisterType.FLAGS or ty == RegisterType.HINT:
            return out
        rep = Armv7mInstruction._build_pattern_replacement(s, ty, arg)
        res = out.replace(f"<{s}>", rep)
        if res == out:
            raise FatalParsingException(f"Failed to replace <{s}> by {rep} in {out}!")
        return res

    @staticmethod
    def _expand_reg_list(reg_list):
        """Expanding list of registers that may contain ranges
        Examples:
        r1,r2,r3
        s1-s7
        r1-r3,r14
        """
        reg_list = reg_list.replace("{", "")
        reg_list = reg_list.replace("}", "")

        reg_list_type = reg_list[0]
        regs = []
        for reg_range in reg_list.split(","):
            if "-" in reg_range:
                start = reg_range.split("-")[0]
                end = reg_range.split("-")[1]
                start = int(start.replace(reg_list_type, ""))
                end = int(end.replace(reg_list_type, ""))
                regs += [f"{reg_list_type}{i}" for i in range(start, end + 1)]
            else:  # not a range, just a register
                regs += [reg_range]
        return reg_list_type, regs

    @staticmethod
    def build_core(obj, res):

        def group_to_attribute(group_name, attr_name, f=None):
            def f_default(x):
                return x

            def group_name_i(i):
                return f"{group_name}{i}"

            if f is None:
                f = f_default
            if group_name in res.keys():
                setattr(obj, attr_name, f(res[group_name]))
            else:
                idxs = [i for i in range(4) if group_name_i(i) in res.keys()]
                if len(idxs) == 0:
                    return
                assert idxs == list(range(len(idxs)))
                setattr(
                    obj, attr_name, list(map(lambda i: f(res[group_name_i(i)]), idxs))
                )

        group_to_attribute("datatype", "datatype", lambda x: x.lower())
        group_to_attribute(
            "imm", "immediate", lambda x: x.replace("#", "")
        )  # Strip '#'
        group_to_attribute("index", "index", int)
        group_to_attribute("flag", "flag")
        group_to_attribute("width", "width")
        group_to_attribute("barrel", "barrel")
        group_to_attribute("label", "label")
        group_to_attribute("reg_list", "reg_list")

        for s, ty in obj.pattern_inputs:
            if ty == RegisterType.FLAGS:
                obj.args_in.append("flags")
            else:
                obj.args_in.append(Armv7mInstruction._to_reg(ty, res[s]))
        for s, ty in obj.pattern_outputs:
            if ty == RegisterType.FLAGS:
                obj.args_out.append("flags")
            else:
                obj.args_out.append(Armv7mInstruction._to_reg(ty, res[s]))

        for s, ty in obj.pattern_in_outs:
            obj.args_in_out.append(Armv7mInstruction._to_reg(ty, res[s]))

    @staticmethod
    def build(c, src):
        pattern = getattr(c, "pattern")
        inputs = getattr(c, "inputs", []).copy()
        outputs = getattr(c, "outputs", []).copy()
        in_outs = getattr(c, "in_outs", []).copy()
        modifies_flags = getattr(c, "modifiesFlags", False)
        depends_on_flags = getattr(c, "dependsOnFlags", False)

        if isinstance(src, str):
            # Leave checking the mnemonic out for now; not strictly required
            # Allows xxx.w and xxx.n syntax
            res = Armv7mInstruction.get_parser(pattern)(src)
        else:
            assert isinstance(src, dict)
            res = src

        obj = c(
            pattern,
            inputs=inputs,
            outputs=outputs,
            in_outs=in_outs,
            modifiesFlags=modifies_flags,
            dependsOnFlags=depends_on_flags,
        )

        Armv7mInstruction.build_core(obj, res)
        return obj

    @classmethod
    def make(cls, src):
        return Armv7mInstruction.build(cls, src)

    def write(self):
        out = self.pattern
        ll = (
            list(zip(self.args_in, self.pattern_inputs))
            + list(zip(self.args_out, self.pattern_outputs))
            + list(zip(self.args_in_out, self.pattern_in_outs))
        )
        for arg, (s, ty) in ll:
            out = Armv7mInstruction._instantiate_pattern(s, ty, arg, out)

        def replace_pattern(txt, attr_name, mnemonic_key, t=None):
            def t_default(x):
                return x

            if t is None:
                t = t_default

            a = getattr(self, attr_name)
            if a is None:
                return txt
            if not isinstance(a, list):
                txt = txt.replace(f"<{mnemonic_key}>", t(a))
                return txt
            for i, v in enumerate(a):
                txt = txt.replace(f"<{mnemonic_key}{i}>", t(v))
            return txt

        out = replace_pattern(out, "immediate", "imm", lambda x: f"#{x}")
        out = replace_pattern(out, "datatype", "dt", lambda x: x.upper())
        out = replace_pattern(out, "flag", "flag")
        out = replace_pattern(out, "index", "index", str)
        out = replace_pattern(out, "width", "width", lambda x: x.lower())
        out = replace_pattern(out, "barrel", "barrel", lambda x: x.lower())
        out = replace_pattern(out, "label", "label")
        out = replace_pattern(out, "reg_list", "reg_list", lambda x: x.lower())

        out = out.replace("\\[", "[")
        out = out.replace("\\]", "]")
        return out


class Armv7mBranch(Armv7mInstruction):
    pass


class Armv7mBasicArithmetic(Armv7mInstruction):
    pass


class Armv7mShiftedArithmetic(Armv7mInstruction):
    pass


class Armv7mMultiplication(Armv7mInstruction):
    pass


class Armv7mLogical(Armv7mInstruction):
    pass


class Armv7mShiftedLogical(Armv7mInstruction):
    pass


class Armv7mLoadInstruction(Armv7mInstruction):
    pass


class Armv7mStoreInstruction(Armv7mInstruction):
    pass


class Armv7mFPInstruction(Armv7mInstruction):
    pass


# FP
class vmov_gpr(Armv7mFPInstruction):
    pattern = "vmov<width> <Rd>, <Sa>"
    inputs = ["Sa"]
    outputs = ["Rd"]


class vmov_gpr2(Armv7mFPInstruction):
    pattern = "vmov<width> <Sd>, <Ra>"
    inputs = ["Ra"]
    outputs = ["Sd"]


class vmov_gpr2_dual(Armv7mFPInstruction):
    pattern = "vmov<width> <Sd1>, <Sd2>, <Ra>, <Rb>"
    inputs = ["Ra", "Rb"]
    outputs = ["Sd1", "Sd2"]

    @classmethod
    def make(cls, src):
        obj = Armv7mInstruction.build(cls, src)
        obj.args_out_combinations = [
            (
                [0, 1],
                [
                    [f"s{i}", f"s{i+1}"]
                    for i in range(
                        0, len(RegisterType.list_registers(RegisterType.FPR))
                    )
                ],
            )
        ]
        return obj


# movs
class movw_imm(Armv7mBasicArithmetic):
    pattern = "movw <Rd>, <imm>"
    outputs = ["Rd"]


class movt_imm(Armv7mBasicArithmetic):
    pattern = "movt <Rd>, <imm>"
    in_outs = ["Rd"]


# Addition
class add(Armv7mBasicArithmetic):
    pattern = "add<width> <Rd>, <Ra>, <Rb>"
    inputs = ["Ra", "Rb"]
    outputs = ["Rd"]


class add_short(Armv7mBasicArithmetic):
    pattern = "add<width> <Rd>, <Ra>"
    inputs = ["Ra"]
    in_outs = ["Rd"]


class add_imm(Armv7mBasicArithmetic):
    pattern = "add<width> <Rd>, <Ra>, <imm>"
    inputs = ["Ra"]
    outputs = ["Rd"]


class add_imm_short(Armv7mBasicArithmetic):
    pattern = "add<width> <Rd>, <imm>"
    in_outs = ["Rd"]


class add_shifted(Armv7mShiftedArithmetic):
    pattern = "add<width> <Rd>, <Ra>, <Rb>, <barrel><imm>"
    inputs = ["Ra", "Rb"]
    outputs = ["Rd"]


class adds(Armv7mBasicArithmetic):
    pattern = "adds<width> <Rd>, <Ra>, <Rb>"
    inputs = ["Ra", "Rb"]
    outputs = ["Rd"]
    modifiesFlags = True


class uadd16(Armv7mBasicArithmetic):
    pattern = "uadd16<width> <Rd>, <Ra>, <Rb>"
    inputs = ["Ra", "Rb"]
    outputs = ["Rd"]


class sadd16(Armv7mBasicArithmetic):
    pattern = "sadd16<width> <Rd>, <Ra>, <Rb>"
    inputs = ["Ra", "Rb"]
    outputs = ["Rd"]


# Subtraction
class sub(Armv7mBasicArithmetic):
    pattern = "sub<width> <Rd>, <Ra>, <Rb>"
    inputs = ["Ra", "Rb"]
    outputs = ["Rd"]


class sub_shifted(Armv7mShiftedArithmetic):
    pattern = "sub<width> <Rd>, <Ra>, <Rb>, <barrel><imm>"
    inputs = ["Ra", "Rb"]
    outputs = ["Rd"]


class sub_short(Armv7mBasicArithmetic):
    pattern = "sub<width> <Rd>, <Ra>"
    inputs = ["Ra"]
    in_outs = ["Rd"]


class sub_imm_short(Armv7mBasicArithmetic):
    pattern = "sub<width> <Ra>, <imm>"
    in_outs = ["Ra"]


class subs_imm(Armv7mBasicArithmetic):
    pattern = "subs<width> <Rd>, <Ra>, <imm>"
    inputs = ["Ra"]
    outputs = ["Rd"]
    modifiesFlags = True


class subs_imm_short(Armv7mBasicArithmetic):
    pattern = "subs<width> <Ra>, <imm>"
    in_outs = ["Ra"]
    modifiesFlags = True


class usub16(Armv7mBasicArithmetic):
    pattern = "usub16<width> <Rd>, <Ra>, <Rb>"
    inputs = ["Ra", "Rb"]
    outputs = ["Rd"]


class ssub16(Armv7mBasicArithmetic):
    pattern = "ssub16<width> <Rd>, <Ra>, <Rb>"
    inputs = ["Ra", "Rb"]
    outputs = ["Rd"]


# Multiplication
class mul(Armv7mMultiplication):
    pattern = "mul<width> <Rd>, <Ra>, <Rb>"
    inputs = ["Ra", "Rb"]
    outputs = ["Rd"]


class mul_short(Armv7mMultiplication):
    pattern = "mul<width> <Rd>, <Ra>"
    inputs = ["Ra"]
    in_outs = ["Rd"]


class mla(Armv7mMultiplication):
    pattern = "mla<width> <Rd>, <Ra>, <Rb>, <Rc>"
    inputs = ["Ra", "Rb", "Rc"]
    outputs = ["Rd"]


class mls(Armv7mMultiplication):
    pattern = "mls<width> <Rd>, <Ra>, <Rb>, <Rc>"
    inputs = ["Ra", "Rb", "Rc"]
    outputs = ["Rd"]


class smulwb(Armv7mMultiplication):
    pattern = "smulwb<width> <Rd>, <Ra>, <Rb>"
    inputs = ["Ra", "Rb"]
    outputs = ["Rd"]


class smulwt(Armv7mMultiplication):
    pattern = "smulwt<width> <Rd>, <Ra>, <Rb>"
    inputs = ["Ra", "Rb"]
    outputs = ["Rd"]


class smultb(Armv7mMultiplication):
    pattern = "smultb<width> <Rd>, <Ra>, <Rb>"
    inputs = ["Ra", "Rb"]
    outputs = ["Rd"]


class smultt(Armv7mMultiplication):
    pattern = "smultt<width> <Rd>, <Ra>, <Rb>"
    inputs = ["Ra", "Rb"]
    outputs = ["Rd"]


class smulbb(Armv7mMultiplication):
    pattern = "smulbb<width> <Rd>, <Ra>, <Rb>"
    inputs = ["Ra", "Rb"]
    outputs = ["Rd"]


class smlabt(Armv7mMultiplication):
    pattern = "smlabt<width> <Rd>, <Ra>, <Rb>, <Rc>"
    inputs = ["Ra", "Rb", "Rc"]
    outputs = ["Rd"]


class smlabb(Armv7mMultiplication):
    pattern = "smlabb<width> <Rd>, <Ra>, <Rb>, <Rc>"
    inputs = ["Ra", "Rb", "Rc"]
    outputs = ["Rd"]


class smlatt(Armv7mMultiplication):
    pattern = "smlatt<width> <Rd>, <Ra>, <Rb>, <Rc>"
    inputs = ["Ra", "Rb", "Rc"]
    outputs = ["Rd"]


class smlatb(Armv7mMultiplication):
    pattern = "smlatb<width> <Rd>, <Ra>, <Rb>, <Rc>"
    inputs = ["Ra", "Rb", "Rc"]
    outputs = ["Rd"]


class smull(Armv7mMultiplication):
    pattern = "smull<width> <Ra>, <Rb>, <Rc>, <Rd>"
    inputs = ["Rc", "Rd"]
    outputs = ["Ra", "Rb"]


class smlal(Armv7mMultiplication):
    pattern = "smlal<width> <Ra>, <Rb>, <Rc>, <Rd>"
    inputs = ["Rc", "Rd"]
    in_outs = ["Ra", "Rb"]


class smlad(Armv7mMultiplication):
    pattern = "smlad<width> <Ra>, <Rb>, <Rc>, <Rd>"
    inputs = ["Rb", "Rc", "Rd"]
    outputs = ["Ra"]


class smladx(Armv7mMultiplication):
    pattern = "smladx<width> <Ra>, <Rb>, <Rc>, <Rd>"
    inputs = ["Rb", "Rc", "Rd"]
    outputs = ["Ra"]


class smmulr(Armv7mMultiplication):
    pattern = "smmulr<width> <Ra>, <Rb>, <Rc>"
    inputs = ["Rb", "Rc"]
    outputs = ["Ra"]


class smuad(Armv7mMultiplication):
    pattern = "smuad<width> <Ra>, <Rb>, <Rc>"
    inputs = ["Rb", "Rc"]
    outputs = ["Ra"]


class smuadx(Armv7mMultiplication):
    pattern = "smuadx<width> <Ra>, <Rb>, <Rc>"
    inputs = ["Rb", "Rc"]
    outputs = ["Ra"]


# Logical


class neg_short(Armv7mLogical):
    pattern = "neg<width> <Rd>, <Ra>"
    inputs = ["Ra"]
    in_outs = ["Rd"]


class log_and(Armv7mLogical):
    pattern = "and<width> <Rd>, <Ra>, <Rb>"
    inputs = ["Ra", "Rb"]
    outputs = ["Rd"]


class log_and_shifted(Armv7mShiftedLogical):
    pattern = "and<width> <Rd>, <Ra>, <Rb>, <barrel><imm>"
    inputs = ["Ra", "Rb"]
    outputs = ["Rd"]


class log_or(Armv7mLogical):
    pattern = "orr<width> <Rd>, <Ra>, <Rb>"
    inputs = ["Ra", "Rb"]
    outputs = ["Rd"]


class log_or_shifted(Armv7mShiftedLogical):
    pattern = "orr<width> <Rd>, <Ra>, <Rb>, <barrel><imm>"
    inputs = ["Ra", "Rb"]
    outputs = ["Rd"]


class eor(Armv7mLogical):
    pattern = "eor<width> <Rd>, <Ra>, <Rb>"
    inputs = ["Ra", "Rb"]
    outputs = ["Rd"]


class eor_short(Armv7mLogical):
    pattern = "eor<width> <Rd>, <Ra>"
    inputs = ["Ra"]
    in_outs = ["Rd"]


class eors(Armv7mLogical):
    pattern = "eors<width> <Rd>, <Ra>, <Rb>"
    inputs = ["Ra", "Rb"]
    outputs = ["Rd"]
    modifiesFlags = True


class eors_short(Armv7mLogical):
    pattern = "eors<width> <Rd>, <Ra>"
    inputs = ["Ra"]
    in_outs = ["Rd"]
    modifiesFlags = True


class eor_shifted(Armv7mShiftedLogical):
    pattern = "eor<width> <Rd>, <Ra>, <Rb>, <barrel><imm>"
    inputs = ["Ra", "Rb"]
    outputs = ["Rd"]

    def write(self):
        self.immediate = simplify(self.immediate)
        return super().write()


class bic(Armv7mLogical):
    pattern = "bic<width> <Rd>, <Ra>, <Rb>"
    inputs = ["Ra", "Rb"]
    outputs = ["Rd"]


class bics(Armv7mLogical):
    pattern = "bics<width> <Rd>, <Ra>, <Rb>"
    inputs = ["Ra", "Rb"]
    outputs = ["Rd"]
    modifiesFlags = True


class bic_shifted(Armv7mShiftedLogical):
    pattern = "bic<width> <Rd>, <Ra>, <Rb>, <barrel><imm>"
    inputs = ["Ra", "Rb"]
    outputs = ["Rd"]


class ubfx_imm(Armv7mLogical):
    pattern = "ubfx<width> <Rd>, <Ra>, <imm0>, <imm1>"
    inputs = ["Ra"]
    outputs = ["Rd"]


class ror(Armv7mLogical):
    pattern = "ror<width> <Rd>, <Ra>, <imm>"
    inputs = ["Ra"]
    outputs = ["Rd"]


class ror_short(Armv7mLogical):
    pattern = "ror<width> <Rd>, <imm>"
    in_outs = ["Rd"]


class rors_short(Armv7mLogical):
    pattern = "rors<width> <Rd>, <imm>"
    in_outs = ["Rd"]
    modifiesFlags = True


class lsl(Armv7mLogical):
    pattern = "lsl<width> <Rd>, <Ra>, <imm>"
    inputs = ["Ra"]
    outputs = ["Rd"]


class asr(Armv7mLogical):
    pattern = "asr<width> <Rd>, <Ra>, <imm>"
    inputs = ["Ra"]
    outputs = ["Rd"]


class asrs(Armv7mLogical):
    pattern = "asrs<width> <Rd>, <Ra>, <imm>"
    inputs = ["Ra"]
    outputs = ["Rd"]
    modifiesFlags = True


class pkhtb(Armv7mShiftedLogical):
    pattern = "pkhtb<width> <Rd>, <Ra>, <Rb>, <barrel><imm>"
    inputs = ["Ra", "Rb"]
    outputs = ["Rd"]


class pkhbt(Armv7mLogical):
    pattern = "pkhbt<width> <Rd>, <Ra>, <Rb>"
    inputs = ["Ra", "Rb"]
    outputs = ["Rd"]


class pkhbt_shifted(Armv7mShiftedLogical):
    pattern = "pkhbt<width> <Rd>, <Ra>, <Rb>, <barrel><imm>"
    inputs = ["Ra", "Rb"]
    outputs = ["Rd"]


# Load
class ldr(Armv7mLoadInstruction):
    pattern = "ldr<width> <Rd>, [<Ra>]"
    inputs = ["Ra"]
    outputs = ["Rd"]

    @classmethod
    def make(cls, src):
        obj = Armv7mInstruction.build(cls, src)
        obj.increment = None
        obj.pre_index = 0
        obj.addr = obj.args_in[0]
        return obj

    def write(self):
        if int(self.pre_index) != 0:
            self.immediate = simplify(self.pre_index)
            self.pattern = ldr_with_imm.pattern
        return super().write()


class ldr_with_imm(Armv7mLoadInstruction):
    pattern = "ldr<width> <Rd>, [<Ra>, <imm>]"
    inputs = ["Ra"]
    outputs = ["Rd"]

    @classmethod
    def make(cls, src):
        obj = Armv7mInstruction.build(cls, src)
        obj.increment = None
        obj.pre_index = obj.immediate
        obj.addr = obj.args_in[0]
        return obj

    def write(self):
        self.immediate = simplify(self.pre_index)

        if self.immediate < 0:
            # if immediate is < 0, the encoding is 32-bit anyway
            # and the .w has no meaning.
            # LLVM complains about the .w in this case
            # TODO: This actually seems to be a bug in LLVM
            self.width = ""
        return super().write()


class ldrb_with_imm(Armv7mLoadInstruction):
    pattern = "ldrb<width> <Rd>, [<Ra>, <imm>]"
    inputs = ["Ra"]
    outputs = ["Rd"]

    @classmethod
    def make(cls, src):
        obj = Armv7mInstruction.build(cls, src)
        obj.increment = None
        obj.pre_index = obj.immediate
        obj.addr = obj.args_in[0]
        return obj

    def write(self):
        self.immediate = simplify(self.pre_index)
        return super().write()


class ldrh_with_imm(Armv7mLoadInstruction):
    pattern = "ldrh<width> <Rd>, [<Ra>, <imm>]"
    inputs = ["Ra"]
    outputs = ["Rd"]

    @classmethod
    def make(cls, src):
        obj = Armv7mInstruction.build(cls, src)
        obj.increment = None
        obj.pre_index = obj.immediate
        obj.addr = obj.args_in[0]
        return obj

    def write(self):
        self.immediate = simplify(self.pre_index)
        return super().write()


class ldr_with_imm_stack(Armv7mLoadInstruction):
    pattern = "ldr<width> <Rd>, [sp, <imm>]"
    inputs = []
    outputs = ["Rd"]

    @classmethod
    def make(cls, src):
        obj = Armv7mInstruction.build(cls, src)
        obj.increment = None
        obj.pre_index = obj.immediate
        obj.addr = "sp"
        return obj

    def write(self):
        self.immediate = simplify(self.pre_index)
        return super().write()


class ldr_with_postinc(Armv7mLoadInstruction):
    pattern = "ldr<width> <Rd>, [<Ra>], <imm>"
    in_outs = ["Ra"]
    outputs = ["Rd"]

    @classmethod
    def make(cls, src):
        obj = Armv7mLoadInstruction.build(cls, src)
        obj.increment = obj.immediate
        obj.pre_index = None
        obj.args_inout_out_different = [(0, 0)]  # Can't have Rd==Ra
        obj.addr = obj.args_in_out[0]
        return obj


class ldrh_with_postinc(Armv7mLoadInstruction):
    pattern = "ldrh<width> <Rd>, [<Ra>], <imm>"
    in_outs = ["Ra"]
    outputs = ["Rd"]

    @classmethod
    def make(cls, src):
        obj = Armv7mLoadInstruction.build(cls, src)
        obj.increment = obj.immediate
        obj.args_inout_out_different = [(0, 0)]  # Can't have Rd==Ra
        obj.pre_index = None
        obj.addr = obj.args_in_out[0]
        return obj


class ldrb_with_postinc(Armv7mLoadInstruction):
    pattern = "ldrb<width> <Rd>, [<Ra>], <imm>"
    in_outs = ["Ra"]
    outputs = ["Rd"]

    @classmethod
    def make(cls, src):
        obj = Armv7mLoadInstruction.build(cls, src)
        obj.increment = obj.immediate
        obj.args_inout_out_different = [(0, 0)]  # Can't have Rd==Ra
        obj.pre_index = None
        obj.addr = obj.args_in_out[0]
        return obj


class Ldrd(Armv7mLoadInstruction):
    pass


class ldrd_imm(Ldrd):
    pattern = "ldrd<width> <Ra>, <Rb>, [<Rc>, <imm>]"
    in_outs = ["Rc"]
    outputs = ["Ra", "Rb"]

    @classmethod
    def make(cls, src):
        obj = Armv7mLoadInstruction.build(cls, src)
        obj.increment = None
        obj.pre_index = obj.immediate
        obj.addr = obj.args_in_out[0]
        return obj

    def write(self):
        self.immediate = simplify(self.pre_index)
        return super().write()


class ldrd_with_postinc(Ldrd):
    pattern = "ldrd<width> <Ra>, <Rb>, [<Rc>], <imm>"
    in_outs = ["Rc"]
    outputs = ["Ra", "Rb"]

    @classmethod
    def make(cls, src):
        obj = Armv7mLoadInstruction.build(cls, src)
        obj.increment = obj.immediate
        obj.pre_index = None
        obj.addr = obj.args_in_out[0]
        return obj


class ldr_with_inc_writeback(Armv7mLoadInstruction):
    pattern = "ldr<width> <Rd>, [<Ra>, <imm>]!"
    in_outs = ["Ra"]
    outputs = ["Rd"]

    @classmethod
    def make(cls, src):
        obj = Armv7mInstruction.build(cls, src)
        obj.increment = obj.immediate
        obj.pre_index = None
        obj.addr = obj.args_in_out[0]
        return obj


class ldm_interval(Armv7mLoadInstruction):
    pattern = "ldm<width> <Ra>, <reg_list>"
    inputs = ["Ra"]
    outputs = []

    def write(self):
        regs = ",".join(self.args_out)
        self.reg_list = f"{{{regs}}}"
        return super().write()

    @classmethod
    def make(cls, src):
        obj = Armv7mLoadInstruction.build(cls, src)
        reg_list_type, reg_list = Armv7mInstruction._expand_reg_list(obj.reg_list)

        obj.args_out = reg_list
        obj.num_out = len(obj.args_out)
        obj.arg_types_out = [RegisterType.GPR] * obj.num_out
        available_regs = RegisterType.list_registers(RegisterType.GPR)
        obj.args_out_combinations = [
            (
                list(range(0, obj.num_out)),
                [list(a) for a in itertools.combinations(available_regs, obj.num_out)],
            )
        ]
        obj.args_out_restrictions = [None for _ in range(obj.num_out)]
        return obj


class ldm_interval_inc_writeback(Armv7mLoadInstruction):
    pattern = "ldm<width> <Ra>!, <reg_list>"
    in_outs = ["Ra"]
    outputs = []

    def write(self):
        regs = ",".join(self.args_out)
        self.reg_list = f"{{{regs}}}"
        return super().write()

    @classmethod
    def make(cls, src):
        obj = Armv7mLoadInstruction.build(cls, src)
        reg_list_type, reg_list = Armv7mInstruction._expand_reg_list(obj.reg_list)

        obj.args_out = reg_list
        obj.num_out = len(obj.args_out)
        obj.arg_types_out = [RegisterType.GPR] * obj.num_out
        obj.increment = obj.num_out * 4

        available_regs = RegisterType.list_registers(RegisterType.GPR)
        obj.args_out_combinations = [
            (
                list(range(0, obj.num_out)),
                [list(a) for a in itertools.combinations(available_regs, obj.num_out)],
            )
        ]
        obj.args_out_restrictions = [None for _ in range(obj.num_out)]
        return obj


class vldr_with_imm(Armv7mLoadInstruction):
    pattern = "vldr<width> <Sd>, [<Ra>, <imm>]"
    inputs = ["Ra"]
    outputs = ["Sd"]

    @classmethod
    def make(cls, src):
        obj = Armv7mInstruction.build(cls, src)
        obj.increment = None
        obj.pre_index = obj.immediate
        obj.addr = obj.args_in[0]
        return obj

    def write(self):
        self.immediate = simplify(self.pre_index)
        return super().write()


class vldr_with_postinc(Armv7mLoadInstruction):
    pattern = "vldr<width> <Sd>, [<Ra>], <imm>"
    in_outs = ["Ra"]
    outputs = ["Sd"]

    @classmethod
    def make(cls, src):
        obj = Armv7mLoadInstruction.build(cls, src)
        obj.increment = obj.immediate
        obj.pre_index = None
        obj.addr = obj.args_in_out[0]
        return obj


class vldm_interval_inc_writeback(Armv7mLoadInstruction):
    pattern = "vldm<width> <Ra>!, <reg_list>"
    in_outs = ["Ra"]
    outputs = []

    def write(self):
        regs = ",".join(self.args_out)
        self.reg_list = f"{{{regs}}}"
        return super().write()

    @classmethod
    def make(cls, src):
        obj = Armv7mLoadInstruction.build(cls, src)
        reg_list_type, reg_list = Armv7mInstruction._expand_reg_list(obj.reg_list)

        obj.args_out = reg_list
        obj.num_out = len(obj.args_out)
        obj.arg_types_out = [RegisterType.FPR] * obj.num_out
        obj.increment = obj.num_out * 4

        available_regs = RegisterType.list_registers(RegisterType.FPR)
        obj.args_out_combinations = [
            (
                list(range(0, obj.num_out)),
                [
                    [f"s{i+j}" for i in range(0, obj.num_out)]
                    for j in range(0, len(available_regs) - obj.num_out)
                ],
            )
        ]
        obj.args_out_restrictions = [None for _ in range(obj.num_out)]
        return obj


# Store


class str_no_off(Armv7mStoreInstruction):
    pattern = "str<width> <Rd>, [<Ra>]"
    inputs = ["Ra", "Rd"]
    outputs = []

    @classmethod
    def make(cls, src):
        obj = Armv7mInstruction.build(cls, src)
        obj.increment = None
        obj.pre_index = 0
        obj.addr = obj.args_in[0]
        return obj

    def write(self):
        if int(self.pre_index) != 0:
            self.immediate = simplify(self.pre_index)
            self.pattern = str_with_imm.pattern
        return super().write()


class strh_with_imm(Armv7mStoreInstruction):
    pattern = "strh<width> <Rd>, [<Ra>, <imm>]"
    inputs = ["Ra", "Rd"]
    outputs = []

    @classmethod
    def make(cls, src):
        obj = Armv7mInstruction.build(cls, src)
        obj.increment = None
        obj.pre_index = obj.immediate
        obj.addr = obj.args_in[0]
        return obj

    def write(self):
        self.immediate = simplify(self.pre_index)
        return super().write()


class str_with_imm(Armv7mStoreInstruction):
    pattern = "str<width> <Rd>, [<Ra>, <imm>]"
    inputs = ["Ra", "Rd"]
    outputs = []

    @classmethod
    def make(cls, src):
        obj = Armv7mInstruction.build(cls, src)
        obj.increment = None
        obj.pre_index = obj.immediate
        obj.addr = obj.args_in[0]
        return obj

    def write(self):
        self.immediate = simplify(self.pre_index)

        if self.immediate < 0:
            # if immediate is < 0, the encoding is 32-bit anyway
            # and the .w has no meaning.
            # LLVM complains about the .w in this case
            # TODO: This actually seems to be a bug in LLVM
            self.width = ""

        return super().write()


class str_with_imm_stack(Armv7mStoreInstruction):
    pattern = "str<width> <Rd>, [sp, <imm>]"
    inputs = ["Rd"]
    outputs = []

    @classmethod
    def make(cls, src):
        obj = Armv7mInstruction.build(cls, src)
        obj.increment = None
        obj.pre_index = obj.immediate
        obj.addr = "sp"
        return obj

    def write(self):
        self.immediate = simplify(self.pre_index)
        return super().write()


class str_with_postinc(Armv7mStoreInstruction):
    pattern = "str<width> <Rd>, [<Ra>], <imm>"
    inputs = ["Rd"]
    in_outs = ["Ra"]

    @classmethod
    def make(cls, src):
        obj = Armv7mStoreInstruction.build(cls, src)
        obj.increment = obj.immediate
        obj.pre_index = None
        obj.addr = obj.args_in_out[0]
        return obj


class strh_with_postinc(Armv7mStoreInstruction):
    pattern = "strh<width> <Rd>, [<Ra>], <imm>"
    inputs = ["Rd"]
    in_outs = ["Ra"]

    @classmethod
    def make(cls, src):
        obj = Armv7mStoreInstruction.build(cls, src)
        obj.increment = obj.immediate
        obj.pre_index = None
        obj.addr = obj.args_in_out[0]
        return obj


class stm_interval_inc_writeback(Armv7mLoadInstruction):
    pattern = "stm<width> <Ra>!, <reg_list>"
    in_outs = ["Ra"]
    outputs = []

    def write(self):
        regs = ",".join(self.args_out)
        self.reg_list = f"{{{regs}}}"
        return super().write()

    @classmethod
    def make(cls, src):
        obj = Armv7mLoadInstruction.build(cls, src)

        reg_list_type, reg_list = Armv7mInstruction._expand_reg_list(obj.reg_list)

        obj.args_in = reg_list
        obj.num_in = len(obj.args_in)
        obj.arg_types_in = [RegisterType.GPR] * obj.num_in
        obj.increment = obj.num_in * 4

        available_regs = RegisterType.list_registers(RegisterType.GPR)
        obj.args_in_combinations = [
            (
                list(range(0, obj.num_in)),
                [
                    [f"s{i+j}" for i in range(0, obj.num_in)]
                    for j in range(0, len(available_regs) - obj.num_in)
                ],
            )
        ]
        obj.args_in_restrictions = [None for _ in range(obj.num_in)]
        return obj


# Other
class cmp(Armv7mBasicArithmetic):
    pattern = "cmp<width> <Ra>, <Rb>"
    inputs = ["Ra", "Rb"]
    modifiesFlags = True
    dependsOnFlags = True


class cmp_imm(Armv7mBasicArithmetic):
    pattern = "cmp<width> <Ra>, <imm>"
    inputs = ["Ra"]
    modifiesFlags = True


class bne(Armv7mBranch):
    pattern = "bne<width> <label>"
    dependsOnFlags = True


class Spill:
    def spill(reg, loc, spill_to_vreg=None, prefix="STACK_LOC"):
        """Generates the instruction text for a spill to either
        the stack or the FPR. If spill_to_vreg is None (default),
        the spill goes to the stack. Otherwise, spill_to_vreg must
        be an integer defining the base of the registers in the FPR
        which should be used as a stack. For example, passing 8 would
        spill to s8,s9,.. ."""
        if spill_to_vreg is None:
            return f"str {reg}, [sp, #{prefix}_{loc}]"
        else:
            vreg_base = int(spill_to_vreg)
            return f"vmov s{vreg_base+int(loc)}, {reg}"

    def restore(reg, loc, spill_to_vreg=None, prefix="STACK_LOC"):
        """Generates the instruction text for a spill restore from either
        the stack or the FPR. If spill_to_vreg is None (default),
        the spill goes to the stack. Otherwise, spill_to_vreg must
        be an integer defining the base of the registers in the FPR
        which should be used as a stack. For example, passing 8 would
        spill to s8,s9,.. ."""
        if spill_to_vreg is None:
            return f"ldr {reg}, [sp, #{prefix}_{loc}]"
        else:
            vreg_base = int(spill_to_vreg)
            return f"vmov {reg}, s{vreg_base+int(loc)}"


def ldm_interval_splitting_cb():
    def core(inst, t, log=None):

        ptr = inst.args_in[0]
        regs = inst.args_out
        width = inst.width

        ldrs = []
        offset = 0
        for r in regs:
            ldr = Armv7mInstruction.build(
                ldr_with_imm, {"width": width, "Rd": r, "Ra": ptr, "imm": f"#{offset}"}
            )
            ldr.pre_index = offset
            ldrs.append(ldr)
            offset += 4

            ldr_src = (
                SourceLine(ldr.write())
                .add_tags(inst.source_line.tags)
                .add_comments(inst.source_line.comments)
            )
            ldr.source_line = ldr_src

        if log is not None:
            log(f"ldm splitting: {t.inst}; {[ldr for ldr in ldrs]}")

        t.changed = True
        t.inst = ldrs
        return True

    return core


ldm_interval.global_fusion_cb = ldm_interval_splitting_cb()


def stm_interval_inc_writeback_splitting_cb():
    def core(inst, t, log=None):

        ptr = inst.args_in_out[0]
        regs = inst.args_in
        width = inst.width

        strs = []
        offset = (len(regs) - 1) * 4
        for r in regs[:0:-1]:
            store = Armv7mInstruction.build(
                str_with_imm, {"width": width, "Rd": r, "Ra": ptr, "imm": f"#{offset}"}
            )
            store.pre_index = offset
            strs.append(store)
            offset -= 4
        # Final store includes increment
        store = Armv7mInstruction.build(
            str_with_postinc,
            {"width": width, "Rd": regs[0], "Ra": ptr, "imm": f"#{len(regs) * 4}"},
        )
        strs.append(store)

        for store in strs:
            store_src = (
                SourceLine(store.write())
                .add_tags(inst.source_line.tags)
                .add_comments(inst.source_line.comments)
            )
            store.source_line = store_src

        if log is not None:
            log(f"stm! splitting: {t.inst}; {[store for store in strs]}")

        t.changed = True
        t.inst = strs
        return True

    return core


stm_interval_inc_writeback.global_fusion_cb = stm_interval_inc_writeback_splitting_cb()


def ldm_interval_inc_writeback_splitting_cb():
    def core(inst, t, log=None):

        ptr = inst.args_in_out[0]
        regs = inst.args_out
        width = inst.width

        ldrs = []
        offset = (len(regs) - 1) * 4
        for r in regs[:0:-1]:
            ldr = Armv7mInstruction.build(
                ldr_with_imm, {"width": width, "Rd": r, "Ra": ptr, "imm": f"#{offset}"}
            )
            ldr.pre_index = offset
            ldrs.append(ldr)
            offset -= 4
        # Final load includes increment
        ldr = Armv7mInstruction.build(
            ldr_with_postinc,
            {"width": width, "Rd": regs[0], "Ra": ptr, "imm": f"#{len(regs) * 4}"},
        )
        ldrs.append(ldr)

        for ldr in ldrs:
            ldr_src = (
                SourceLine(ldr.write())
                .add_tags(inst.source_line.tags)
                .add_comments(inst.source_line.comments)
            )
            ldr.source_line = ldr_src

        if log is not None:
            log(f"ldm! splitting: {t.inst}; {[ldr for ldr in ldrs]}")

        t.changed = True
        t.inst = ldrs
        return True

    return core


ldm_interval_inc_writeback.global_fusion_cb = ldm_interval_inc_writeback_splitting_cb()


def vldm_interval_inc_writeback_splitting_cb():
    def core(inst, t, log=None):

        ptr = inst.args_in_out[0]
        regs = inst.args_out
        width = inst.width

        ldrs = []
        offset = 0
        for r in regs:
            ldr = Armv7mInstruction.build(
                vldr_with_imm, {"width": width, "Sd": r, "Ra": ptr, "imm": f"#{offset}"}
            )
            ldr.pre_index = offset
            ldrs.append(ldr)
            offset += 4

        add_ptr = Armv7mInstruction.build(
            add_imm, {"width": width, "Rd": ptr, "Ra": ptr, "imm": f"#{offset}"}
        )
        ldrs.append(add_ptr)

        for ldr in ldrs:
            ldr_src = (
                SourceLine(ldr.write())
                .add_tags(inst.source_line.tags)
                .add_comments(inst.source_line.comments)
            )
            ldr.source_line = ldr_src

        if log is not None:
            log(f"ldm! splitting: {t.inst}; {[ldr for ldr in ldrs]}")

        t.changed = True
        t.inst = ldrs
        return True

    return core


vldm_interval_inc_writeback.global_fusion_cb = (
    vldm_interval_inc_writeback_splitting_cb()
)


def ldrd_postinc_splitting_cb():
    def core(inst, t, log=None):

        ptr = inst.args_in_out[0]
        regs = inst.args_out
        width = inst.width

        ldrs = []

        ldr = Armv7mInstruction.build(
            ldr_with_imm, {"width": width, "Rd": regs[1], "Ra": ptr, "imm": "#4"}
        )
        ldr.pre_index = 4
        ldrs.append(ldr)
        # Final load includes increment
        ldr = Armv7mInstruction.build(
            ldr_with_postinc, {"width": width, "Rd": regs[0], "Ra": ptr, "imm": "#8"}
        )
        ldr.increment = 8
        ldr.pre_index = None
        ldr.addr = ptr
        ldrs.append(ldr)

        for ldr in ldrs:
            ldr_src = (
                SourceLine(ldr.write())
                .add_tags(inst.source_line.tags)
                .add_comments(inst.source_line.comments)
            )
            ldr.source_line = ldr_src

        if log is not None:
            log(f"ldrd splitting: {t.inst}; {[ldr for ldr in ldrs]}")

        t.changed = True
        t.inst = ldrs
        return True

    return core


ldrd_with_postinc.global_fusion_cb = ldrd_postinc_splitting_cb()


def ldrd_imm_splitting_cb():
    def core(inst, t, log=None):

        ptr = inst.args_in_out[0]
        regs = inst.args_out
        width = inst.width

        ldrs = []

        ldr = Armv7mInstruction.build(
            ldr_with_imm,
            {"width": width, "Rd": regs[0], "Ra": ptr, "imm": inst.pre_index},
        )
        ldr.pre_index = inst.pre_index
        ldrs.append(ldr)
        # Final load includes increment
        ldr = Armv7mInstruction.build(
            ldr_with_imm,
            {"width": width, "Rd": regs[1], "Ra": ptr, "imm": f"{inst.pre_index}+4"},
        )
        ldr.pre_index = f"{inst.pre_index}+4"
        ldr.addr = ptr
        ldrs.append(ldr)

        for ldr in ldrs:
            ldr_src = (
                SourceLine(ldr.write())
                .add_tags(inst.source_line.tags)
                .add_comments(inst.source_line.comments)
            )
            ldr.source_line = ldr_src

        if log is not None:
            log(f"ldrd splitting: {t.inst}; {[ldr for ldr in ldrs]}")

        t.changed = True
        t.inst = ldrs
        return True

    return core


ldrd_imm.global_fusion_cb = ldrd_imm_splitting_cb()


# Returns the list of all subclasses of a class which don't have
# subclasses themselves
def all_subclass_leaves(c):

    def has_subclasses(cl):
        return len(cl.__subclasses__()) > 0

    def is_leaf(c):
        return not has_subclasses(c)

    def all_subclass_leaves_core(leaf_lst, todo_lst):
        leaf_lst += filter(is_leaf, todo_lst)
        todo_lst = [
            csub
            for c in filter(has_subclasses, todo_lst)
            for csub in c.__subclasses__()
        ]
        if len(todo_lst) == 0:
            return leaf_lst
        return all_subclass_leaves_core(leaf_lst, todo_lst)

    return all_subclass_leaves_core([], [c])


Instruction.all_subclass_leaves = all_subclass_leaves(Instruction)


def iter_armv7m_instructions():
    yield from all_subclass_leaves(Instruction)


def find_class(src):
    for inst_class in iter_armv7m_instructions():
        if isinstance(src, inst_class):
            return inst_class
    raise UnknownInstruction(
        f"Couldn't find instruction class for {src} (type {type(src)})"
    )



"""
Experimental Cortex-M4 microarchitecture model for SLOTHY

WARNING: The data in this module is approximate and may contain errors.
"""

# ################################ NOTE #############################################
#                                                                                   #
#  WARNING: The data in this module is approximate and may contain errors.          #
#           They are _NOT_ an official software optimization guide for Cortex-M4.   #
#                                                                                   #
# ###################################################################################

from enum import Enum
from slothy.helper import lookup_multidict
from slothy.targets.arm_v7m.arch_v7m import (
    find_class,
    ldr_with_imm,
    str_with_imm,
    adds,
    add,
    add_short,
    add_imm,
    add_imm_short,
    mul,
    smull,
    smlal,
    log_and,
    log_or,
    eor,
    bic,
    sub_imm_short,
    ror,
    add_shifted,
    sub_shifted,
    eor_shifted,
    bic_shifted,
)

issue_rate = 1
llvm_mca_target = "cortex-m4"


class ExecutionUnit(Enum):
    """Enumeration of execution units in Cortex-M4 model"""

    UNIT = 0

    def __repr__(self):
        return self.name


# Opaque function called by SLOTHY to add further microarchitecture-
# specific constraints which are not encapsulated by the general framework.
def add_further_constraints(slothy):
    if slothy.config.constraints.functional_only:
        return


# Opaque function called by SLOTHY to add further microarchitecture-
# specific objectives.
def has_min_max_objective(config):
    """Adds Cortex-"""
    _ = config
    return False


def get_min_max_objective(slothy):
    _ = slothy
    return


execution_units = {
    (
        adds,
        add,
        add_short,
        add_imm,
        add_imm_short,
        add_shifted,
        sub_shifted,
        sub_imm_short,
        mul,
        smull,
        smlal,
        log_and,
        log_or,
        eor,
        eor_shifted,
        bic,
        bic_shifted,
        ror,
        ldr_with_imm,
        str_with_imm,
    ): ExecutionUnit.UNIT,
}

inverse_throughput = {
    (
        adds,
        add,
        add_short,
        add_imm,
        add_imm_short,
        add_shifted,
        sub_shifted,
        sub_imm_short,
        mul,
        smull,
        smlal,
        log_and,
        log_or,
        eor,
        eor_shifted,
        bic,
        bic_shifted,
        ror,
    ): 1,
    (ldr_with_imm, str_with_imm): 2,
}

default_latencies = {
    (
        adds,
        add,
        add_short,
        add_imm,
        add_imm_short,
        add_shifted,
        sub_shifted,
        sub_imm_short,
        mul,
        smull,
        smlal,
        log_and,
        log_or,
        eor,
        eor_shifted,
        bic,
        bic_shifted,
        ror,
    ): 1,
    (ldr_with_imm, str_with_imm): 2,
}


def get_latency(src, out_idx, dst):
    _ = out_idx  # out_idx unused
    instclass_src = find_class(src)
    latency = lookup_multidict(default_latencies, src, instclass_src)

    return latency


def get_units(src):
    instclass_src = find_class(src)
    units = lookup_multidict(execution_units, src, instclass_src)
    if isinstance(units, list):
        return units
    return [units]


def get_inverse_throughput(src):
    instclass_src = find_class(src)
    return lookup_multidict(inverse_throughput, src, instclass_src)



class FatalParsingException(Exception):
    """A fatal error happened during instruction parsing"""


class UnknownInstruction(Exception):
    """The parent instruction class for the given object could not be found"""


class UnknownRegister(Exception):
    """The register could not be found"""
